#!/bin/sh
# the next line restarts using wish \
exec wish "$0" "$@"

# tk#.tcl --
#
#      A simple spreadsheet especially designed to show off the
#      properties of tcl/tk.
#
# Copyright (c) 1997-1999 GPL / Martin Vermeer & The tk# Team
# mv@fgi.fi
#
# Click "About" for info on copying and lack of guarantees :-)
#
# Change log:
#
#    MV 06-Sep-97 Base version
#    MV 08-Sep-97 Changes mainly in ReDisplay to improve performance
#    MV 10-Sep-97 Changes due to remarks by Bruce Gingery:
#                 Quit panel added "cancel"
#                 Colour choice cancel error fixed
#                 Current cell uses textbg as fg for clarity
#                 Clear added to Format menu
#                 Selected link paint in on-line manual
#                 Manual language improvement + Quickstart addition
#    MV 11-Sep-97 Fixed watch cursor bug in rename row/col
#                 Fixed multi-word left align format bug
#                 Corrected manual explanation of [Sum ...] etc. functions
#                 Added tagging of selected rows/columns
#    MV 12-Sep-97 Added select match string construction for aggregates
#                 Fixed watch cursor bug Select Box
#                 Fixed current cell hilite bug Recalc-Sel
#    MV 14-Sep-97 Added functions AboveCell...RightCell; Today, USAtoday.
#                 Removed set "Modified" flag if cell content unchanged.
#                 Template variable $i packed in parentheses to prevent
#                 match between, e.g., "Row1" and "Row10".
#                 Added REPOSITORIES for function and other definitions,
#                 as suggested by Bruce Gingery to support extensibility.
#    MV 15-Sep-97 Called clock control buttons Start and Stop.
#                 Multiple cmd line load filenames implemented
#                 Export comma separated table added; removal of separator
#                 from field values.
#                 Create directory tl# in user root for repository files.
#    MV 16-Sep-97 Fixed bug in NewNameError: used "string first" function.
#                 Bug in FormattedCell: test existence v-element
#    MV 17-Sep-97 Final bug fixes. Redefine $initdir as [pwd] if dir
#                 nonexistent (other machine/user). 
#                 Default ext .tl# for load repository.
#                 Scrollbar for listboxes 
#    -----------  Launch 0.1
#    MV 19-Sep-97 Default ext .dat offered in Export/Import box
#    MV 20-Sep-97 Content Match added (formulae or data)
#    MV 21-Sep-97 Added "unformat selection" and "un-expr selection".
#    MV 23-Sep-97 Sanitized format and expr statements by introducing
#                 procs "Fmt" and "Expr". Demos must be revised now
#                 (they still work)
#                 Separate "left align" / "left align text" 
#                 Import linear, newline separated list of cells (crude!)
#    MV 24-Sep-97 Bug fixes. ^V for content search on data.
#                 Control-Mouse1: goto cell without entering value into
#                 Formula box.
#    MV 27-Sep-97 Got newline separated list import properly working.
#    MV 29-Sep-97 Added "Make Sheet" cmd for initializing no. of rows/cols
#                 (thanks Juergen Vigna)
#                 Added "formula stack" (thanks Peter da Silva)  
#                 Tried to fix recursive ERR 4 problem (?)
#    MV 30-Sep-97 Asks for spreadsheet size on startup if no datafile
#                 name on cmd line.
#
#    MV 04-Oct-97 Perfected "Add Row": take name from Formula: -box.
#                 Used "catch" for error in regexp string from Formula:
#                 -box for content match.
#                 $exp now saved/loaded.
#                 sort option for row/col move.
#                 Row/col rename by clicking on titlerow.
#                 Case sensitive/insensitive option for content match.
#                 Lyx output (more or less) working.
#    MV 05-Oct-97 Now working well. Cells containing \ are latexed
#                 (allows handling of diacriticals, or even embedded
#                 LaTeX such as \parbox for multiline cells)
#    MV 06-Oct-97 Cleanup in Row/Col Spawn routines.
#    MV 07-Oct-97 Now Row/Col Spawn/Move/Kill are cleaned up so their
#                 display part does not unnecessarily build the text
#                 widget content from scratch again. VERY fast!
#                 May have introduced some bugs, but seems to work.
#    MV 09-Oct-97 Bug fixes (thanks Jurgen Vigna!). 
#                 Resize width on column selection added.
#                 Replace info text by checkbuttons & direction syms.
#                 Make Ctrl-Mouse1 the cmd to move cell to formula box.
#    MV 10-Oct-97 Added <Del> = goto empty formulabox feature (Add Row). 
#                 Font cleanup etc. in manual. "Exit Help" button.
#    MV 13-Oct-97 Further speedup by removing large row no. bottlenecks
#                 (lsearch, lappend large $row list). ReDisplay speedup.
#                 Benjamin Allan inspired this. Parentheses bug in Geometry
#                 fixed (now <$i> used for rowname generation).
#    MV 14-Oct-97 Added entry of row & col ranges: [RR r1 r2] for row range
#                 Fixed problems of manual exit into loaded file;
#                 (thanks Juergen Vigna)
#    MV 15-Oct-97 Reorganized aggregate functions menus. 
#                 Lyx 0.12 tables work.
#    MV 16-Oct-97 Grey out menus when in help. Bug fixes. Lyx col alignment.
#                 QuitWarn debugging :-(
#                 Fixed ColWidth speed (text widget refresh)
#    MV 17-Oct.97 Allow col/row name change to its own sub/superstring 
#                 (NewNameError optional 3rd arg).
#                 Made "Kill Sheet" and "Kill Select" a little safer :-/
#                 Content Search using selection by current cell in title
#    MV 20-Oct-97 Precise match on title names in "Matchmaker" and in 
#                 NewNameError. Better than "containment match".
#    -----------  Launch 0.2 
#    MV 28-Oct-97 Proc "Now" changed to allow sorting on date string.
#    MV 14-Nov-97 Make Export use the name stem of $filename for
#                 $exportname. More intuitive.
#    MV 08-Dec-97 Added pie chart plot with PostScript output.
#    MV 30-Mar-98 Made repositories reside in either /etc/tl#/ or
#                 ~/tl#/
#    MV 28-Aug-98 Added Export button.
#                 Print Selection -menu item added on request.
#                 Based on pipe. "|Pipe Selection".
#                 Also Print All: "|Pipe All".
#                 Same functionality incorporated in Export: Write "|lpr -h" 
#                 for filename.
#    MV 18-Aug-99 Cleaned up in import/export of "csv" files. Import 
#                 respects now the "titlecol" switch.
#    MV 20-Aug-99 Added a "label print" option. The label format menu is made
#                 a repository. Two label sizes defined. The output is a
#                 collection of TeX boxes in a LaTeX wrapper. Works almost OK.
#    MV 25-Aug-99 Cleaned up the "label print" logic. Now uses $labels(top) to
#                 distinguish among simple mail merge, sticky labels and
#                 BibTeX database output. BibTeX support added -- wasn't that
#                 hard.
#    -----------  Launch 0.3
#
#    MV 18-Oct-04 Replaced -com by -command in a number of places.
#                 Fixed the thing with inserting "\n" on rendering. Apparently
#                 inserting at end does it.
#    MV 26-Jan-05 Call this 0.4
#    -----------  Launch 0.4

# init --
#
# proc to initialize a number of things needed to be able to display
# an empty worksheet, and continue working.
#
proc init {argc argv} {
global thisformula formulastack r c rows cols headrows sidecols f v cw 
global rowsel colsel
global autocalc modified filename exportname inspoint colours initdir
global reload tr tc defmov find winsize env tcl_precision reps
global fonts paintallowed exp printcommand
global labels

# User configurable

    set winsize "550x355+0+50"
    set printcommand "enscript -h"

# end user config

    set tcl_precision 17

    set autocalc 1

    array set find [list case 0 fv "Data" rowsel 0 colsel 0]
    array set exp [list howmuch "All" titlerow 0 titlecol 0 \
              align "Align Right" separ "Tab"]

    # to detect multiple (overlaying) loads/imports:
    set reload   0 

    TemplateDefault
    set rows {}
    set cols {}
    set sidecols {title}
    set headrows {title}
    set r title
    set c title
    foreach x {title} {set cw($x) 12; set cs($x) 0} 
    foreach x {title} {set rs($x) 0} 
    set thisformula ""
    foreach i {1 2 3 4 5} {set formulastack($i) ""}
    set f(title,title) ""
    set rowsel {}; set colsel {}
    set inspoint {^$}
    # Create a dedicated data directory. 
    if ![file exists "$env(HOME)/tk#"] {
        if [catch {exec mkdir "$env(HOME)/tk#"} result] {
            puts $result
        }
    } 
    # Create a directory for user function repositories.
    if ![file exists "$env(HOME)/tl#"] {
        if [catch {exec mkdir "$env(HOME)/tl#"} result] {
            puts $result
        }
    }
    ColourDefault

    set filename "datafile.tk#"
    set exportname ""
    set initdir [pwd]
    set defmov "-"
    set reps {}

    set paintallowed 1

    array set fonts [list\
        man    "*-helvetica-medium-r-*-*-12*"\
        exit   "*-helvetica-bold-r-*-*-12*"\
        title  "*-helvetica-*-r-*-*-24*"\
        icon   "*-times-*-i-*-*-24*"\
        header "*-helvetica-*-r-*-*-18*"\
        mono   "*-courier-medium-r-*-*-12*"]

     array set labels [list\
        top    ""\
	start  ""\
        itemstart   ""\
        itemend   ""\
	end    "\n"\
	bottom ""]
    
    Getrc
    return
}

# ColourDefault --
#
#      Set the colours to acceptable default values. 
#      Will be overridden by any loaded files.
#
proc ColourDefault {} {
global colours
    # set up colours:
    # textfg textbg      for text in panels
    # textcur            current cell bg
    # selcolour          selected rows/cols
    # inscolour          insert col/row position
    # timecolor          timer (clock) cell     
    # iconcolour         tk# icon colour
    # mancolour          used for titles in manual
    set colourlist {textfg black textbg   white                \
                    textcur      blue     selcolour     yellow \
                    inscolour    red      timecolour    green  \
                    iconcolour   blue     mancolour     blue }
    foreach {x y} $colourlist {
        set colours($x) $y
    }
    return
}

# TemplateDefault --
#
#      Reset "template" to empty string and "spawn" to 1.
#
proc TemplateDefault {} {
global template spawn
    set spawn 1
    set template ""
}

# InitIndexRow --
#
# Initialize table rs of row starting indices. Return no. of rows.
#
# Arguments:
#      prows   Row name list to be processed in this call.
#              Either $rows or $headrows.
#
proc InitIndexRow {prows} {
global rs
    set bgn 0
    foreach r $prows {
        array set rs "$r $bgn"
        incr bgn
    }
    return $bgn
} 

# InitIndexCol --
#
#      Initialize tables cs of col starting indices and cw of col widths.
#      Return total no. of character positions in a row. 
#
# Arguments:
#      pcols   Col name list to be processed in this call.
#              Either $cols or $sidecols.
#
proc InitIndexCol {pcols} {
global cs cw
    set bgn 0
    foreach c $pcols {
        set cs($c) $bgn
        if ![info exists cw($c)] {set cw($c) 12}
        incr bgn $cw($c)
        }
    return $bgn
}

# InitIndex --
#
#      Init the row and col start and width tables for all panels.
#
proc InitIndex {} {
global rows cols headrows sidecols
global cw cs rs rowsel colsel f
    InitIndexRow $headrows
    InitIndexRow $rows
    InitIndexCol $sidecols
    InitIndexCol $cols
}

# PanelsState --
#
#      Set state to "normal" or "disabled" 
#      to make spreadsheet panels uneditable.
#
proc PanelsState {state} {
global frq
    foreach panel {corner side head main} {
        $frq.$panel config -state $state
    }
}

# PutIcon --
#
#      Place an icon on the corner panel for About and Manual
#
# Argument:
#      w     Intended width of panels "corner" and "side"
#
proc PutIcon {w} {
global colours fr frq  r c tools b
global fonts paintallowed
    set paintallowed 0
    MenusState disabled

    $frq.corner config -width $w -height 7
    $frq.head config -height 7
    $frq.side config -width $w
    $frq.corner delete 1.0 end

    # Construct the "tk#" icon from textual elements:
    
    $frq.corner tag configure tktag -foreground $colours(mancolour) \
                             -background $colours(textbg) -justify center
    $frq.corner tag configure hashtag -offset -12 \
                             -foreground $colours(iconcolour) \
                             -background $colours(textbg) -just center
# test
#    puts [catch {$frq.corner tag configure tktag -font $fonts(icon)}] 
#    puts [font actual $fonts(icon)]

    # Handle possible non-existence of times 34 or helv 24:
    if [catch {$frq.corner tag configure tktag -font $fonts(icon)}] {
        if [catch {$frq.corner tag configure tktag -font $fonts(title)}] {
            $frq.corner tag configure tktag -font $fonts(header)
            $frq.corner tag configure hashtag -font $fonts(header)
        } else {$frq.corner tag configure hashtag -font $fonts(title) }
    } else {$frq.corner tag configure hashtag -font $fonts(icon) }

    $frq.corner insert end "tk" tktag
    $frq.corner insert end "#\n" hashtag

    $frq.head delete 1.0 end
    $frq.head tag configure titletag -font $fonts(title) -underline true\
                             -foreground $colours(iconcolour) \
                             -background $colours(textbg) -justify center
    $frq.head tag configure subtitletag -font $fonts(man)\
                             -foreground $colours(mancolour) \
                             -background $colours(textbg) -justify center

    $frq.head tag configure exittag -justify center -borderwidth 2\
            -font $fonts(exit) -background #d9d9d9 -relief raised
    $frq.head insert end {### (tk-number) ###} titletag
    $frq.head insert end {
A spreadsheet with a difference!
Copyright (c) 1997-1999 GPL
Martin Vermeer & The tk# Team.
    } subtitletag
    $frq.head insert end " Back to Worksheet " exittag
    $frq.head tag bind exittag <Button-1> {
        ReDisplay
        PutInfo
        # Trick; see GetMouse1
        set paintallowed -1
        MenusState normal
    }
    return
}
# MenusState --
#
#      Change the state of the menus and buttons from normal to
#      disabled and back, when displaying help
#
#
proc MenusState {state} {
global tools b fr
    foreach x {add start calc find} {
        $tools.$x config -state $state
    }
    foreach x {file editrc format sel calc math colour clock} {
        $b.$x config -state $state
    }
    foreach x {l e ew} {
        $fr.box1.$x  config -state $state
    }
    return
}

# About --
#
#      Put the "About..." information into the spreadsheet.
#     
proc About {} {
global frq fonts

    PanelsState normal
    PutIcon 11
    $frq.side delete 1.0 end
    $frq.main delete 1.0 end
    $frq.main tag configure centered -justify center -font $fonts(man)
    $frq.main insert end {
No promises, no lies; may be freely copied and 
improved in agreement with the spirit of the 
Gnu General Public License, the text of which is 
pro forma included by this reference.

Version 0.3 
} centered
    PanelsState disabled
    return
}

# HandleButton1 --
#
#      On-line manual support routine
#      Show the chapter heading for this link;
#      highlight chosen link in index panel
#
# Argument:
#      tagname   the tag to be made visible in the main panel.
#
proc HandleButton1 {tagname} {
global frq colours fonts
    # Only way to get rid of old highlights: Redraw it all :-(
    Manual
    $frq.main see $tagname.first
    $frq.side see $tagname.first
    $frq.side tag configure $tagname -background $colours(mancolour) \
              -foreground $colours(textbg) 
    return
}

# PutManChapter --
#
#      Put a single manual chapter, with title, properly tagged, into
#      both the side panel (titles only) and the main panel (titles + text).
#
# Arguments:
#      tagno     Tag label, only used internally to connect panels "side" and
#                "main".
#      text      The name of the chapter = link title appearing in "side".
#      chapter   The plain text of the chapter (or part of it).
#
proc PutManChapter {tagno text chapter} {
global frq colours fonts 
    # creates a button in the "side" panel, carrying a keyword
    # which is searched for (command!) in the "main" panel if pressed.

    $frq.side tag configure t$tagno -foreground $colours(mancolour)\
              -background $colours(textbg) -font $fonts(man) -underline true
    $frq.main tag configure t$tagno -font $fonts(header)\
              -foreground $colours(mancolour)\
              -background $colours(textbg) -justify center
    
    $frq.main tag configure tt -font $fonts(man) -tabs 1c

    # intelligent TOC: clicking mouse scrolls text to visibility.
    $frq.side tag bind t$tagno <Button-1> [list HandleButton1 t$tagno]
    
    $frq.side insert end $text t$tagno
    $frq.side insert end \n
    $frq.main insert end $text t$tagno 

    $frq.main insert end $chapter\n tt
    return
}

# PutTaggedText --
#
#      Put an amount of manual text with a tag connected to it 
#      into main panel.
#
# Arguments:
#      t      Tag to be used.
#      text   Text in "main" panel to apply it to.
#
proc PutTaggedText {t text} {
    .w.fr.frq.main insert end $text $t
    return
}

# PutText --
#
#      Put an amount of manual text as such to the main panel.
#
# Argument:
#      text   The text to put into "main" panel.
#
proc PutText {text} {
global frq
    $frq.main insert end $text tt
}

# Manual --
#
#      Put the on-line manual text to the main panel screen.
#
proc Manual {} {
global frq global colours fonts 

    PanelsState normal
    PutIcon 24

    $frq.side delete 1.0 end

    $frq.main delete 1.0 end
    $frq.main configure -wrap word

    # Define the formatting tags we may want to use:
    $frq.main tag config tabindent -lmargin2 2c -tabs 1c -font $fonts(man)
    $frq.main tag config mono     -font $fonts(mono)
    $frq.main tag config titletag -font $fonts(title) -underline true\
              -foreground $colours(iconcolour) -justify center

    $frq.main insert end {
tk# on-line manual

} titletag

    PutManChapter 1 Introduction {
tk# is a simple but intelligent spreadsheet written in tcl/tk.

The graphic abilities of tk are used to provide a grid of four\
interconnected, scrolling panels containing cells with formulas\
and data values associated with them:}
PutTaggedText mono {
                       <-------------> 
         +-------------+-------------+
         !   corner    !     head    !
      ^  +-------------+-------------+
      !  !    side     !     main    !
      v  +-------------+-------------+
}
   PutText {
The top and left panels ("head" and "side"), which are narrow\
and scroll only in one direction, contain the names of the rows/columns. 

Additionally, they may contain user defined things such as row and column\
sums, so-called "aggregates", which then also remain always visible even\
when navigating through the (large) main panel.

The scrollable main panel, on the other hand, contains the bulk\
of the data. Its scrolling behaviour helps when entering large\
amounts of data. 

Because of performance reasons, you should not make the main panel larger\
than, say, 100 x 100 cells; you do wiser to split your application into\
smaller parts handled separately, save selections from them, and load these\
together. Be aware that tk# allows saving of the selection (with row/column\
titles) as well as multiple loading of data files with overlaying, cf. below.

One thing that makes tk# special is that IT DOES NOT USE NUMBERS but names\
to refer to rows and columns. Also selection of ranges is based\
on names and regular expression matching, as is the evaluation of\
formulas over cell ranges (summation etc.) which is an important\
part of real life spreadsheet usage (think of bookkeeping!).

}
PutManChapter 1a Quickstart {
Start up tk# without a filename, and choose 10 rows and 10 columns.

This will set up a 10 x 10 spreadsheet blank to get you started.\
Just enter values, strings and expressions (see below) into the\
Formula: -box, when it has the focus (ESC toggles). Pressing\
ENTER will put them into the sheet.

Alternatively: Start up tk#, and give the command

	Edit/Geometry
	No. of rows:					10
	No. of Columns:					10

and click OK. Proceed to enter formulas as above.

If you want to control what row and column names are generated,\
give the following commands in succession:

	Edit/Row Spawn
	Spawn Factor:					10
	Template:					R$i

	Edit/Col Spawn
	Spawn Factor:					10
	Template:					C$i
}
    PutManChapter 2 Functions {
Within expressions, you can use the full math facility,\
including the function library, provided with tcl; see the\
tcl manual. Also all the common operators are provided.

Formulas should be tcl expressions: one round of variable and command\
substitution is done. Please note that numerical expressions must be\
enclosed in

	[expr ... ]

in order to be evaluated. The ellipsis represents your expression, e.g.

	[expr sqrt(2)]

to get the square root of 2. This enclosing is achieved by pressing\
control-E in the Formula: entry box. Alternatively, use the Format menu\
item.

Strings, bot numeric and non-numeric, are copied straight over to\
the corresponding cell value. This is useful for text and\
direct data entries.
}

    PutManChapter 2a {Aggregate Functions} {
There is a set of SPECIAL or AGGREGATE FUNCTIONS implemented in tk# for\
summation and the like.

NOTE. In these "summation" functions, always summation over the CURRENT\
column (c) or row (r) is implied. The second argument is a MATCH STRING,\
which is matched against either row or column NAMES. This is a non-trivial\
thing, quite different from traditional spreadsheets, and should be well\
understood. }
    PutTaggedText tabindent {
	[Sum c June]
		sums elements in THIS column and in ALL rows\
containing the string "June" in their names.

	[Sum r Sales]
		row sum over all cols with "Sales" in name

	[Av  c June]
		average

	[Cnt r Rent]
		number of columns named Rent-something

	[Max c Point]
		E.g. maximum coordinate value (in this column) \
of point set

	[Min c Point]
		Minimum same.

	[wSum c June wtscol]
		column sum of June rows weighted\
by entries in the column "wtscol".

		This one can be generalized by defining,\
in the column "wtscol" (e.g. the neighbouring column), the variable\
$wtscol by the statement "set wtscol $r". 

		After that, writing

			[wSum c June $wtscol]

		will give the same sum as above. However,\
now RENAMING the weights column to something else will not break the sum!

	[wAv c June $wtsrow]
		weighted average.

	[Ssq  r Gas]
		sum of squares

	[wSsq r Gas $wtscol]
		same, weighted

	[RMS  c June]
		root-mean-square

	[wRMS c June $wtscol]
		weighted

	[Sd  r Rent]
		standard deviation

	[wSd r Rent]
		weighted 
}
    PutText {
These last functions are not very well tested yet.

}
    PutManChapter 2aa {About Regular Expressions} {
You get an aggregate function through the math menu. It will look like

	[Sum r _]

with "_" indicating the cursor. You can then manually enter a string\
that will be used as e REGULAR EXPRESSION mathing COLUMN names.

Alternatively, if you choose the column alternative:

	[Sum c _]

you are exprected to enter a regex matching ROW names. 

Read the tcl manual or online help "man n regexp" for further\
explanation.

NOTE the use of square brackets forcing, what is called\
"command substitution" i.e. execution of the procedure enclosed.

Example of use: }
    PutTaggedText mono {
title     csums    |-spalte1    -spalte2 
rsums              |[Sum c sei] [Sum c sei]
-------------------+-----------------------
seile1- [Sum r spa]|         12          18 
seile2- [Sum r spa]|         15           3 
seile3- [Sum r spa]|          7           8 
}
    PutText {
Upon recalc you should see the sums appear. The argument "sei" inside\
the [ ] causes a match with the row names containing this substring, in this\
case all three of them.

NOTE NOTE NOTE (you MUST UNDERSTAND this!):
As you see from the example, when summing through a ROW (i.e. [Sum r ...]\
you should give a match string for COLUMN headers ("spa").\
And vice versa!

NOTE also that it pays off to choose row and column names judiciously,\
so you can use them easily in this way!

A facility that makes selecting rows/columns for summation easier,\
is the tagging/untagging facility. These are found in the Row/Col Rename\
subwindows. Use them by using the match string argument "#$" in your\
summation command, e.g.

	[Sum c #$]

}
    PutManChapter 2b {"Matchmaker"} {
Check out also the "matchmaker": It generates a string for including all\
selected rows/columns in a summation operation. As follows: }
    PutTaggedText tabindent {
	*	Select the range you want;
	*	Write the summation formula you need into the Formula: -box,\
with the cursor prepositioned;
	*	Choose "matchmaker" from the menu.
}
    PutText {
After that, you have a formula that will sum over all rows/cols you\
selected above. 

E.g. if you want to do a column summation over all\
elements IN THE CURRENT COLUMN on rows "seile1", "seile2", "seile3", \
you have to compose the following regular expression formula ("alternation"): 

	[Sum c seile1|seile2|seile3]
}
    PutText {
This is precisely what "matchmaker" does for you.

Then, select the cells you want to move this formula into; there are\
several equivalent ways, but perhaps the easiest is just to click\
with the left mouse on the top left, and then with the right mouse\
on the bottom right cell of your target interval.

After that, move focus back to the formula entry box (with the mouse,\
or with the ESC key) and press ENTER. Voila!

A less wasteful alternative, manual way to get the same sum, is the formula 

	[Sum c sei\[1-3\]]

NOTE: pay attention to the use of \[ instead of [, and \]\
instead of ]. The backslash ("escape character") is needed to get the\
square brackets -- indicating a character class -- past the tcl\
interpreter -- which would treat it as a command substitution indicator,\
i.e. it would try to EXECUTE the command "1-3" -- into the\
regex matching engine!
}
    PutText {
This is an example of using a REGULAR EXPRESSION.\
Note also that, in this case, as we want\
ALL rows, we can simply use "sei" as the matching regular expression,\
as done in the above example table.

}
    PutManChapter 2bb "Cell Ranges" {
Another useful feature are row and column ranges. You can enter a row or column\
range function into a formula through the Math menu. Proceed as follows:}
    PutTaggedText tabindent {
	* 	Select an "aggregate function", e.g. Column Sum.\
This will put 

			[Sum c _] 

		into the Formula: -box, with "_" representing\
the cursor.
	*	Select "Row Range"; result:  

			[Sum c [RR _]]

	*	Put the cursor on the starting cell of the range, and\
press "Current Row Name". Result: 

			[Sum c [RR row1 _]]

	*	Put the cursor on the ending cell and press again\
"Current Row Name". Result: 

			[Sum c [RR row1 row2]]

	*	Now you can select a target cell or selection to put the\
formula in, move focus to the Formula: - box, and press ENTER.\
The value of the range sum should appear in the cell.
}
    PutText {
All these summation methods require the cells to be in rows or columns.\
summing irregularly located cells can only be done directly, by entering\
the formula (e.g.):

		[expr $v(salararies,tot)+$v(taxes,tot)+$v(rawmat,tot)]

}
    PutManChapter 3 "Selecting, Finding Cells" {
The most direct way to select cells is by using the mouse. A cell range\
thus selected is always CONTIGUOUS.

First you select the cells to edit by putting an "anchor" using\
mouse button 1 (the left one). This selects this one\
cell, making it the CURRENT CELL). The anchor will be top left of\
the selected range.

Mouse-3 (the right one) then marks the end (bottom right) of a selected range,\
both in columns and in rows. The range is painted on the screen,\
but ONLY in the row and column headers (this is for efficiency reasons\
with large spreadsheets).

As an alternative, on can directly select a NON-CONTIGUOUS set of\
row or column names in listboxes, accessible through the\
Select/Row Select Listbox or Select/Column Select Listbox menu items.\
While a thus selected range may be non-contiguous, it will of course\
always be an intersection of a row and a column selection, cf. below.

tk# offers also intelligent, "mouse-free" selection alternatives: One can use\
regular expression matching either or row or column names (titles) or on\
spreadsheet content!

Alternative Three: You can use the Select by Title Match -item from the Select\
menu, to give two "match strings" to select row and column names\
with. NOTE that a select range defined in this way may be non-contiguous.

The matching procedure used is regular expression matching (REGEXP,\
see tcl manual "man n regexp").

Alternative Four: Content Match. All rows\
and columns are selected in which any of the cells contain the match string.\

}

    PutManChapter 3s "Content Match" {
Content match can be done on the formulas or on the values in the cells.\
It is your choice. You can also choose "Case Sensitive", which means that\
a case-precise check is done; or not, in which case upper and lower case\
letters are treated as equivalent. Matching can be made case-free by\
unsetting the checkbox in Options.

Note that, when matching values, the value matched is the variable\
$v($r,$c). This variable (cell) begins at the cell edge with something\
that matches ^, and ends at an edge that matches $.

Generally the whole sheet it searched by this command. However, you can\
force a search of only the row selection (and all columns!) by putting\
the current cell cursor in the title row. Similarly you can force a\
search of the column selection (and all rows) by putting the current\
cell in the title column on the left.
}
    PutManChapter 3a "Focus transfer" {
Here we give a summary of the mouse and keyboard commands used to transfer\
focus from the spreadsheet panels to the "Formula:" editing box, and some\
other things.}

    PutTaggedText tabindent {
	Mouse-1:
		Set current cell; set selection to current cell only.

	Mouse-3:
		Extend selection contiguously (in row and columns)\
to this cell.

	Control-Mouse1 / Control-Enter:
		Set current cell; set selection to it;\
MOVE current cell formula to "Formula:" -box for editing.

	ESC:
		Change focus between EDIT and NAV; do NOT change current cell\
or selection.

	ENTER:
		Changes the selection to become the current cell only. Does\
not change focus.

	DEL:
		When in NAV mode, goes to EDIT focus and empties "Formula:"\
-box. Its content is scrolled up, out of view, but recoverable.

	Ctrl-DEL:
		Same as above, but also adjusts selection to\
be only the current cell. To fill consecutive cells quickly,\
set the default movement to "Right", and type

 "Ctr-Del (cell contents) Return Ctrl-Del (cell contents) Return"

etcetera. The "Add" button puts you into this entry mode.
}
    PutText {
You should try these alternatives out to get a feeling\
for the mouse buttons and key combinations behaviour!

NOTE (technical remark) that a selected range is ALWAYS the intersection\
of a row selection and a column selection. Inside the software, they\
are represented by the lists $rowsel and $colsel.

}
    PutManChapter 4 "Editing Cells" {
Once you have made your selection, and moved in into the Formula:-\
box as described above, you can start editing it there.

If you pressed "control-left mouse button" on a cell, it already moved focus\
to the Formula: -box. This is the quickest way for editing only\
one cell.

You can also select ONE CELL by using the keyboard only: use the\
arrow keys to go to the cell, then press Control-ENTER to get it into the\
Formula: -box. 

If you press ESC, the Formula: -box will remain unchanged. In fact, you can \
toggle between NAV and EDIT modes by just pressing ESC. The focus will change\
correspondingly between the spreadsheet panels and the "Formula:" edit box.

Useful for entering lots of data is pressing Ctrl-DEL when in NAV mode. This\
will change focus to EDIT mode and empty the Formula:- box for your\
entry. Any previous content is scrolled up. Use this with the "Add"\
button to create new rows (put the new row name into the Formula: -box\
and press Add).

Note that the Formula: title is a menu button: click to make the items\
visible. The Formula: box contains a STACK of five elements, that you\
can scroll through with Control-Up and Control-Down. Use Control-Del\
for quick emptying.

EXPERIMENT!
} 
    PutManChapter 5 "Formula Entry" {
When you are finished editing (but still in EDIT focus! Press ESC if\
you are not), pressing ENTER moves your formula (expression)\
into the spreadsheet. Ctrl-DEL cancels your expression; ESC just moves focus\
to the worksheet, leaving your expression intact.

If a select RANGE is active (not just one cell) your entry will\
be moved into EVERY cell in the range. You can include counters\
$i, $j (row, column within select area starting from 0) or row,\
column names ($r, $c) to make the entries different. You can\
use this to fill in a large number of cells at once and\
corresponds a little bit with "relative cell numbering" in\
other spreadsheets.

If you put the current cell upon a title name (row or column title),\
by clicking Mouse-1 on it, then entering a new name in the Formula: box\
and pressing ENTER will change the name of this row/column.\
This way of renaming is similar to the menu command (no equal names\
allowed etc.)
}
    PutManChapter 6 "Formula Formatting" {
When focus is on the Formula: -box, you can give commands to format\
or unformat the entry in the box. The commands are all given in the\
Format menu, and have control key shortcuts. }
    PutTaggedText tabindent {
	^F	Fixed real
		requires a Format entry in the entrybox right, e.g. 12.6

	^G	Floating real
		same here

	^N	Integer	
		requires format e.g. 8

	Left align
		for data. No blanks allowed.

	^L	Left align text
		for multi-word texts

	^C	Currency
		two digits fixed point.

	^E	Add Expression
		enclose in [Expr...]

	Remove Expr
		the reverse operation

	Remove format
		removes the [Fmt...] strings added by the\
above formatting commands.

	Act on Selection
		Open a submenu for performing the same operations\
collectively not on the current formula box content, but on the\
whole SELECTION.  Powerful command, watch out!

	Replace by value
		IRREVERSIBLE!
}
    PutText {
What these menu commands do, as do the equivalent control sequences, is\
placing a 

	[Fmt "%{fmt}f" "%f" ...] 

thing around the expression to be formatted.\
This operation is reversible, as the menu shows. The value "fmt" is taken\
from the "Format:" box and is of form

	12.6

for real values, and 

	12

for integers. Just like in FORTRAN (a stone age programming language that\
some of you might remember) or in tcl!

Furthermore, you can change the display column width by directly\
editing the "Colwidth:" entry box top right, and pressing ENTER.

}
    PutManChapter 7 "The Insert Mark" {
The insert mark is set on either the row or column header (the\
first row or column containing their names) by clicking the\
middle mouse button (Mouse-2) on it.

Alternatively, Select/Insert Point gives you a listbox to choose from.

Conventionally, if the insert point is lacking, it is assumed\
AFTER the LAST row/column. Otherwise all insertions take place\
before the marked row or column, moving the ones beyond aside to\
make room.

You can also insert rows or columns into the HEADER OR SIDE PANEL.\
In order to do this, put the insert mark on "title" in the top left\
corner. If you then "spawn" rows or columns (or move existing ones)\
they will go into these non-main panels.

Cells in other-than main panels should ONLY be used to contain things\
like row or column titles, as well as accumulation results (subtotals)\
etc. Typically things you want to be always visible when scrolling\
the main panel.
}
    PutManChapter 8 "Manipulating Rows & Cols" {
The Edit menu provides for a number of things to do with rows,\
columns, groups of these and selected areas. } 
    PutTaggedText tabindent {
	Spawn:	
		Take a group of rows/cols (the selection)\
and copy it to the place indicated by the\
insert marker.\
This command asks for a TEMPLATE to\
generate the new row/col names. This\
template may contain the counter $i starting from 0. 

	Example:     
		Spawn counter					3
		Template					Row[expr $i + 1]

	will insert three new rows under the names Row1, Row2, Row3. 

	Kill:
		Delete one or more rows/columns.\
The selection (row or col) is deleted.

	Move:
		Move selection or rows/cols into the\
insert point location.

	Move & Sort:
		Move and SORT row or col selection.

	Rename:
		As the name indicates. Takes a row/col\
template which may contain the counter $i.

		An alternative, more straightforward way to rename a \
row or column is to click on the row or column name header,\
enter the new name in the "Formula:" -box and press ENTER.
}
    PutText {
A special case of rename is TAGGING: this\
adds the suffix "#" (hash), which can be subsequently used in operations\
involving row or column name string matching. Such operations are\
especially the row/col summation ("aggregate") operations discussed earlier.

In spawning and renaming, a new name is not accepted and the operation\
cancelled if the name is a substring of an already existing name,\
or an already existing name is a substring of the new name.\
This would namely interfere with the use of the "matchmaker"\
feature in aggregate operations.

}
    PutManChapter 9 "Special user functions" {
You can put a CLOCK ticking in a cell. Mainly for keeping\
hour sheets at the job. Moving to another cell and actuating\
will stop and "value-ize" the current cell, and start the\
new one ticking.

The format of the clock value, which becomes visible if you do\
a ReCalc or StopClock operation, is taken (1) from the Format:\
-box, and this being empty, taken to be ColWidth.2. [requires work]

If you try to quit with the clock still running, it will be stopped,\
the file marked modified, and a save proposed to you.  }
    PutTaggedText tabindent {
	CellAbove, CellBelow, CellLeft and CellRight:
		produce as function values (i.e. [CellAbove] etc...) the data values in these\
neighbour cells. Useful for e.g. displaying a cumulative sum.

	Now, Today, USAtoday:
		produces the current date and time, or current date only,\
or date in USA style format.

}
    PutManChapter 10 Extensibility {
tk# is EXTENSIBLE: You can enter procedure definitions as\
formulas into cells by just writing them into the formula box,\
and after recalculation they will be available for use.

This is the extensibility intended for OLEO, here implemented\
by the code bootstrap property (similar to LISP) of tcl.

Another product attempting at similar extensibility is SIAG\
(Scheme In A Grid) based on another LISP dialect.

Contrary to those attempts, tk# is not only extensible, it also\
borrows its special character from the programming & extension\
language used, tcl/tk. This may take a little getting used to; but\
once you have, you will discover for yourself that you can do things\
with tk# that you wouldn't even have thought of with a traditional\
spreadsheet!
}
    PutManChapter 11 "Loading and Saving" {
tk# saves the data file "intelligently" as a tcl/tk file to be sourced.\
This applies for a complete as well as a partial (select range) save.

Multiple loads on top of each other are supported;\
then an intelligent merge takes place, keeping the union of\
row/column names in memory. }
    PutTaggedText tabindent {
	Save:
		to same file

	Save As:
		to another file. Prompts for filename.

	Save Select:
		same, but only select range (plus row, col titles)\
is saved.

        Pipe Selection (e.g. print):
	        a VERY powerful feature. You put a command in the\
formula box, and the current selection will be pumped\
into the pipe given! E.g. write "lpr -h" and the selection will go to\
the printer, without a header page. 

	Import:
		Works only with sheets exported WITH headers, \
using tab, comma or newline as separators. No LaTeX or LyX.

	Export:
		Supports output of ASCII tables with\
various field (cell) separators:\
tab, blank, newline, LaTeX-style, LyX. 

		Supports output of full sheet,\
selection, selection plus row and/or column headers.

		LaTeX format can be inserted directly\
into a LaTeX file.

		LyX format is for Version 0.10 or 0.12 tables.\
Import proceeds by loading the\
file generated INTO ITS OWN LYX WINDOW (it is a complete, minimal LyX\
file), and copying the table over to your working file window.

		Tables may contain LaTeX code, in which case\
the whole cell is painted LaTeX red inside LyX.

		Newline separated files are useable as adress\
lists for the LaTeX mailmerge facility "textmerg". 

		Also, if you use the\
		labels.tl# REPOSITORY, you can output sheets of mailing\
		labels as a LaTeX output file.
 
		Using labels.tl# it is also possible to output a BibTeX\
		bibliography database. In this case you have to name the\
		columns according to the fields allowed in a BibTeX file\
		(i.e. author, title, journal, address, ...), and the rows\
		should be named according to the reference types allowed,\
                a colon and the label of the reference\
		(i.e. Book_Vermeer99, phdthesis:jones86, ...;\
		the label will be split off). 
		
		Outputting to a "file" of form "|lpr -h" will pipe to printer\
		etc.

	Quit:
		Terminate tk#, but save/query if modified.

	Plot:
		Plot a pie chart, bar chart or line graph.\
The values are taken from the second selected column, the category names\
from the first selected column. Only values on selected rows are plotted.
}
    PutText {
The use of "plot" requires that a REPOSITORY named plot.tl# is loaded.\
See "Repository" elsewhere in this manual.

You can give file name(s) on the command line. tk# will load them\
in succession, overlaying them. A wildcard filename expands to a list\
the elements of which are filenames loaded successively.

The tk# icon doubles as an indicator for "Modified" (i.e. red if modified,\
green if not) and as a button for saving.

}
    PutManChapter 11a "Repository" {
A special word should be said about REPOSITORIES of function, binding\
and widget definitions. It is possible to write a file containing such\
definitions, which is then automatically loaded whenever an application\
data file is loaded. These function etc. definitions then exist for the user,\
but are NOT bound to any cell and also are NOT re-installed every time\
the sheet is re-calculated.

The way to associate a function repository with a data file is the following:\
}
    PutTaggedText tabindent {
1.		Load the data file (default extension .tk#) using\
the "File/Load" menu command;
2.		Load the repository file using the\
"File/Load Repository" menu command;
3.		Save the data file.
}
  PutText {
This repository idea originated from Bruce Gingery, in support of\
 extensibility (and it works like a dream... a golden idea).

Note that tk# assumes the repository file always to be in the tl# directory\
under /etc, or the tl# directory under your personal home. These are scanned\
in this order.

Now, when re-loading this file, you can use the menu command\
"File/Show Repositories" to display the associated repository file(s) in\
the window title row. 

The actual WRITING of repository files presupposes fluency in tcl/tk and a\
sound understanding of the internal workings of tk#. It is not for beginners.

}
    PutManChapter 12 Variables {
The variables listed below are available for inclusion in formulas.

This means, you can put a formula in a cell\
(enclosed by [expr...] -brackets) which operates on other cells' \
contents. E.g. add up two cells' contents into a third one.\
When adding RANGES of cells, you should however use the\
AGGREGATE operations already discussed earlier. }
    PutTaggedText tabindent {
	$r $c
		row, col number

	$i $j
		numerical row/col counter used in\
Spawn statements and in multiple\
cell editing. Counts from 0 upward.

	$v($r,$c)
		cell values. The array index $r,$c consists of the cell\
row and column names. 

	$w
		column $c width.

NOTE:		You should NOT refer to cell values using\
literal row/column names directly. If you\
rename a row or column, any reference to it will break.

	Rather, put a statement

		[set usecol $c]

	in the column you intend to use,\
and refer to $usecol after that.

}
    PutManChapter 13 "File Names" {
Internally, the program tk# uses three filenames:}
    PutTaggedText tabindent {
	$filename
		The current filename used when saving;\
this should always be displayed in the window header.

	$lfl
		The Last File Loaded.

	$lfs
		The Last File Saved.
    }
    PutText {
The filenames $lfl and $lfs are displayed in the File menu and can be\
loaded by clicking upon. Note that these two filenames are always\
different.

The file names $filename, $lfl and $lfs are NOT saved to the data file.\
It is however possible to put them MANUALLY there. This is a trick,\
which may be used e.g. to load a template file and then save the data\
to a file, the name of which is generated according to some rule\
specified by you. E.g. if you put in the data (template) file the\
command 

    set filename "$env(HOME)/tk#/[Now].tk#"

the template file will be loaded, but $filename set to a date and time\
string, which will become the name of the next saved data file!\
The procedures "Now", "Today" and "USAtoday"are defined in tk#.\
You can define your own, either in tk# or in your source-able data file\
(repository file).

An advanced trick, to be used with considerable care...

}
    PutManChapter 14 "Limitations" {
Many. I have tried to avoid "feature creep", this is more like a\
tool to be used as a building block (UNIX philosophy).

Somewhat optimized for performance with large number of rows\
(suggestion Benjamin Allan). Same optimization for many\
columns not done; would be possible.

The code and its comments should be improved and extended and made to\
conform to Ray Johnson's Tcl Style Guide, so also others can participate\
in development (MV :-)).
}

    PutManChapter 15 "Finally..." {
tk# is distributed under the name tknum, to cater for the needs of\
Netscape and (possibly) other mail/ftp agents. You should rename it on\
your disk to "tk#", if your OS will allow it :-), which is its proper\
name.

Happy tcl/tk-ing!                    Martin Vermeer mv@fgi.fi
    }
    PanelsState disabled
    NavFocus
}

#################################
##
## High level procedures operating on rows or columns:
##

# Iter --
#
#      General service routine for iterative user functions (sum, average, etc.)
#      globalize anything that might be used in $start, $step and $final:
#
# Arguments:
#
# startScript   script to be executed once when starting
# stepScript    script to be executed iteratively
# finalScript   script to be executed once when finishing
# match   match string for row/cols to be included in iteration
# rr1     first row/col list
# rr2     second row/col list
# rr      r for row-wise, c for col-wise
#
# Returns:
#        error messages in return values.
#
proc Iter {startScript stepScript finalScript match rr1 rr2 rr} {
global r c v 
    # ALWAYS output something after a trapped error, NEVER leave catch dangling!
    if [catch {eval $startScript} result] {return "?ERR 1 $result"}
    if {$rr == "r"} {
        set oldr $r ; # save...
        foreach r [concat $rr1 $rr2] {
            if {[regexp -- $match $r]} {
                if [catch {eval $stepScript} result] {return "?ERR 8 $result"}
            }
        }
        set r $oldr ; # ...restore
    } else {
        set oldc $c
        foreach c [concat $rr1 $rr2] {
            if {[regexp -- $match $c]} {
                if [catch {eval $stepScript} result] {return "?ERR 9 $result"}
            }
        set c $oldc
        }
    }
    # REMEMBER that, if $finalscript contains return, catch returns 2!
    if {[catch {eval $finalScript} result] == 1} {return "?ERR 2 $result"}
    return $s ; # This is the $s defined in SumAv!
}

# Citer --
#
#      Wrapper proc for iteration through a column (i.e. over rows).
#
# Arguments:
#      startScript stepScript finalScript  See header of proc Iter.
#      match                               Matching string for selecting
#                                          rownames of cells to include.
#
proc Citer {startScript stepScript finalScript match} {
global headrows rows
    return [Iter $startScript $stepScript $finalScript $match $headrows $rows r] 
}

# Riter --
#
#      Wrapper proc for iteration through a row (i.e. over columns).
#
#      startScript stepScript finalScript  See header of proc Iter.
#      match                               Matching string for selecting
#                                          colnames of cells to include.
#
proc Riter {startScript stepScript finalScript  match} {
global sidecols cols
    return [Iter $startScript $stepScript $finalScript $match $sidecols $cols c] 
}

# -MakeIterCmd --
#
#      Switching routine to direct general specific iteration routines
#      to either Citer or Riter, dpending on $rc flag.
#
# Argument:
#      rc     r means: call Riter, c means: call Citer. 
#
proc MakeIterCmd {rc} {
    if {$rc == "c"} {
        return "Citer"
    } elseif {$rc == "r"} {
        return "Riter"
    } else {error "bad row/col flag: $rc"}
}

# SumAv --
#
#      Generalized routine used by all the others for aggregate operations.
#      
# Arguments:
#      rc      r (row) or c (col)
#      match   match string with row or col names to select elements to be
#              aggregated
#      sa      s (sum) or a (average). Flag for post-division by sum of
#              weights
#      wn      w (use weights) or n (do not)
#      wt      name of row/col providing weights
#      pwr     "power" parameter: 1 for ordinary sum/average, 2 for sum of
#              squares 
#
# Results:
#      Function value.
#      Error messages go to spreadsheet cell or are "puts" in background
#      window (warnings).
#
proc SumAv {rc match sa wn {wt 0} {pwr 1}} {
    # Process the Sum/Average flag:
    if {$sa == "a"} {
        set final {
            if [catch {set s [expr $s / $w]} result] {set s "?ERR 3 $result"}
            }
    } elseif {$sa == "s"} {set final {}} else {error "bad sum/av flag: $sa"}

    # Process the Weighted/Nonweighted flag:
    if {$wn == "w"} {
        if {$rc == "r"} {        
            set stepwt {
                if [set ww 0; catch {scan $v($wt,$c) "%f" ww} result] {
                    puts "?ERR 6 $result"
                } 
                set w [expr $w + $ww]
            } 
        } else { 
            set stepwt {
                if [set ww 0; catch {scan $v($r,$wt) "%f" ww} result] {
                    puts "?ERR 7 $result"
                } 
                set w [expr $w + $ww]
            }
        }
    } elseif {$wn == "n"} {
        set stepwt {set ww 1; incr w}
    } else {error "bad weighting flag: $wn"}

    # Initialize the (possible) weights column:
    set initwt "set wt $wt"
    # Initialize the "step" code:
    switch -- $pwr {
    1 {set stepdo { ; set s [expr $s + $ww * $vv] }}
    2 {set stepdo { ; set s [expr $s + $ww * $vv * $vv] }}
    }
    # Next, either Citer or Riter is executed:
    set s [[MakeIterCmd $rc] [concat {
            # start
            set s 0
            set w 0
            set ww 1 ;
        } $initwt] [concat {
            # step
            # if you can't find a float, put zero:
            if [catch {set vv 0; scan $v($r,$c) "%f" vv} result] {
                # puts "?WRN 1 $result"
            } ;
        } $stepwt $stepdo] $final $match]

    # Return as function value of SumAv:
    return $s
}

# Sum
proc Sum {rc match} {
    return [SumAv $rc $match s n]
}

# Average
proc Av {rc match} {
    return [SumAv $rc $match a n]
}

# Count cells
proc Cnt {rc match} {
    return [[MakeIterCmd $rc] {set s 0} {incr s} {return $s} $match]
}

# Maximum
proc Max {rc match} {
    return [[MakeIterCmd $rc] {set s 0} {
        if {$v($r,$c) > $s} {set s $v($r,$c)}
    } {return $s} $match]
}

# Minimum
proc Min {rc match} {
    return [[MakeIterCmd $rc] {set s 0} {
        if {$v($r,$c) < $s} {set s $v($r,$c)}
    } {return $s} $match]
}

# Weighted sum
proc wSum {rc match wt} {
    return [SumAv $rc $match s w $wt] 
}

# Weighted average
proc wAv {rc match wt} {
    return [SumAv $rc $match a w $wt]
}

# Sum of squares
proc Ssq {rc match} {
    return [SumAv $rc $match s n 0 2]
}

# Weighted sum of squares
proc wSsq {rc match wt} {
    return [SumAv $rc $match s w $wt 2] 
}

# Average of squares
proc MS {rc match} {
    return [SumAv $rc $match a n 0 2]
}

# Root Mean Square
proc RMS {rc match} {
    return [expr sqrt([MS $rc $match])]
}

# Weighted average of squares:
proc wMS {rc match wt} {
    return [SumAv $rc $match a w $wt 2] 
}

# Root of Weighted average of squares:
proc wRMS {rc match wt} {
    return [expr sqrt([wMS $rc $match $wt])] 
}

# Standard deviation:
proc Sd {rc match} {
    set av [Av $rc $match]
    return [expr sqrt([MS $rc $match] - $av * $av)]
}

# Weighted:
proc wSd {rc match wt} {
    set wav [wAv $rc $match $wt]
    return [expr sqrt([wMS $rc $match $wt] - $wav * $wav)]
}

# Colour --
#
#      Colour choice interaction wrapper routine.
#      Typically the arg and function value are elements
#      of the "Colours" array.
#
# Argument:
#      colour       present value of colour.
#      colourname   name of this colour
#
# Result:
#      New colour value in function value.
proc Colour {colour colourname} {
    set colour2 [tk_chooseColor -initialcolor $colour \
        -title "tk# colour choice: $colourname"]
    if {$colour2 != ""} {
        return $colour2
    } else {
        return $colour
    }
}

# Load --
#
#      Loads by "sourcing" a tk# data file, generated by "Save".
#      All vars that one wants to make available to the program should be
#      included in the "globals" list.
#
# Arguments:
#      sw      Flag indicating if 1 that filename should be entered manually.
#              If 0, last filename is used. 
#
proc Load {sw} {
# These are all variables that, if loaded, will be globally available:
global rows cols headrows sidecols f cw r c tr tc
global colours initdir reload lfl lfs
global rowsel colsel inspoint filename winsize
global defmov exp env reps autocalc

# for use in tl# files:
global b tools e ew ef

    .w config -cursor watch
    if {$sw == 2} {
        set typelist {
            {"tl# files" {".tl#"}}
            {"All files" {*}} 
        }
    } else {
        set typelist {
            {"tk# files" {".tk#"}} 
            {"All files" {*}} 
        }
    }
    if {$sw > 0} {
    # Perhaps we changed machine/userid :-)
    if ![file isdirectory $initdir] {set initdir [pwd]}
        set fn [tk_getOpenFile -defaultextension ".tk#" -initialdir $initdir \
            -filetypes $typelist -initialfile $filename -title "tk# file load"]
    } else {set fn $filename}

    if {($fn != "") && [file exists $fn]} {
        if {$sw < 2} {set filename $fn}
        wm title .w "Load: $fn"
        # Last file loaded:
        if {$filename != $lfs} {set lfl $filename}

        # copy row/col lists over:
        foreach x {rows cols headrows sidecols} {set ${x}2 [subst \$$x]}

        source $fn

        # merge row/col lists, merging common names (overwrite!):
        foreach x {rows cols headrows sidecols} {
            foreach y [subst \$$x] {
                if {([lsearch [subst \$${x}2] $y] == -1)} {
                    lappend ${x}2 $y
                }
            }
            set $x [subst \$${x}2]
        }
        # Last colour merged in has it!
        ColoursSet
        # Always one recalc to get values visible!
        ReCalc
        if $reload Modified else {set reload 1}

        if {$sw == 2} {
            # Add to lib stack if called with $sw == 2:
	    lappend reps [file tail $fn]
        } else {
            # Source the library stack:
            if [info exists reps] {
                foreach x $reps {
		    if [file exists /etc/tl#/$x] {source /etc/tl#/$x}
		    if [file exists $env(HOME)/tl#/$x] {
		         source $env(HOME)/tl#/$x
		    }
		}
            }
            wm geometry .w $winsize 
        }
        wm title .w "File: $filename"

    }
    .w config -cursor {} 
    return
}

# ClipRC --
#
#      Return only those rows/cols in $rr which are in the selection $rs.
#
# Arguments:
#      rr     row/col name list, e.g. $rows or $sidecols.
#      rs     row/col selection list, i.e. $rowsel or $colsel
#
# Result:
#      Function value containing list of elements of $rr also in $rs.
#
proc ClipRC {rr rs} {
    set cc {}
    foreach x $rr {
        if {!([lsearch $rs $x] == -1)} {lappend cc $x}
    }
    return $cc 
}

# Save --
#
#      Saves the spreadsheet and its state vars to a data file in "sourceable"
#      format. The globals list should contain everything that one wants
#      saved.
#
# Arguments:
#      sw2      Switch flagging manual entry of filename (> 0) or not ( = 0).
#               If =2, then save only selection.
#
proc Save {sw2} {
# These are all variables that will be saved:
global rows cols headrows sidecols r c tr tc
global colours initdir lfs lfl winsize
global rowsel colsel inspoint f cw filename 
global autocalc defmov exp reps

    set winsize [wm geometry .w]
    # Condition necessary so Save does not break in destroyed window:
    if [winfo exists .w] {.w config -cursor watch}
    set typelist {
        {"tk# files" {".tk#"}} 
        {"All files" {*}} 
    }
    if {$sw2 > 0} {
    if ![file isdirectory $initdir] {set initdir [pwd]}
        set fnx [tk_getSaveFile -defaultextension ".tk#" -initialdir $initdir \
                 -initialfile $filename -filetypes $typelist \
                 -title "tk# file save"]
    } else {set fnx $filename} 

    if {$fnx != ""} {
        set filename $fnx
        if {$filename != $lfl} {set lfs $filename ;# last file saved}

        if {$sw2 == 2} {
            # save the selection only, i.e. restrict the $row and $col lists
            # only to include the selection (+title):
            set sidecols [concat title [ClipRC $sidecols $colsel]]
            set cols [ClipRC $cols $colsel]
            set headrows [concat title [ClipRC $headrows $rowsel]]
            set rows [ClipRC $rows $rowsel]
        } 
        # not saved:
        unset typelist sw2 
        if [winfo exists .w] {wm title .w "Save: $filename"}
        set FileId [open $filename w]

        # Put aside into local vars, prevent saving of originals:
        foreach {lfs2 lfl2 fnx} [list $lfs $lfl $filename] {break}
        # Prevent from being saved, as their unintended loading is wrong:
        unset filename lfl lfs    ; # prevent saving

        # First simple variables...
        foreach x [info vars *]    {
            if ![array exists $x] {
                # A variable may be unset, but still exist:
                if [info exists $x] {
                    puts $FileId [list set $x [subst \$$x] ]
                }
            }
        }
        # ...then arrays cw and f:
        # (Note: elems no longer ref'ed to in rows/cols lists are NOT saved)
        foreach y [concat $sidecols $cols] {
            puts $FileId [list array set cw [array get cw $y]]
        }
        foreach x [ConcatNoTitle $headrows $rows] {
            foreach y [ConcatNoTitle $sidecols $cols] {
                puts $FileId "array set f \{$x,$y \{[set f($x,$y)]\}\}"

            }
        }
        # ... and colours, exp:
        foreach y {colours exp} {
            foreach x [array names $y] {
                puts $FileId [list array set $y [array get $y $x]]
            }
        }
        close $FileId

        # Restore:
        foreach {lfs lfl filename} [list $lfs2 $lfl2 $fnx] {break}

        UnModified
        if [winfo exists .w] {wm title .w "File: $filename"}
    }
    if [winfo exists .w] {.w config -cursor {}} 
    return
}

# Get the "filename stack" file from user home
proc Getrc {} {
global lfl lfs env
    set fn "$env(HOME)/.tk#files"
    if [file exists $fn] {
        set FileId [open $fn r]
        # lfl: Last File Loaded
        gets $FileId lfl
        # lfs: Last File Saved
        gets $FileId lfs 
        close $FileId
    } else {
        set lfl "" 
        set lfs ""
    }
    return
}

# Write back filename stack file to user home:
proc Exit {} {
global lfl lfs env 
    set FileId [open "$env(HOME)/.tk#files" w]
    puts $FileId $lfl
    puts $FileId $lfs 
    close $FileId
    exit
}

# QuitWarn --
#
#      Quit with warning to save file
#      Stop clock if running
#
#      WARNING: Do not change unless you are VERY WELL
#               at home with tcl/tk callback functions!
#               Look at all QuitWarn calls and their
#               interaction (took me 2 hrs to get right!?)
#
# Result:
#      Yes:     Save if modified and exit
#      No:      Do not save; exit
#      Cancel:  Do not even exit. Return to spreadsheet.
#
# Argument:
#      sw:      1 if emergency stop
#
proc QuitWarn {{sw 0}} {
global modified filename tr
    # If clock running, stop it
    if [info exists tr] StopClock

    # Emergency save:
    if {$sw == 1} {
        set messageboxtype yesno
    } else {
        set messageboxtype yesnocancel
    }

    # Has file been modified?
    if $modified {set choice \
            [tk_messageBox -type $messageboxtype -icon question \
                -message "Save File\n $filename?" -default yes]
        if {$choice == "yes"} {
            Save 0
            Exit
        } 
        if {$choice == "no"} {
            Exit
        }
    } else {
       Exit
    }
    return
}

# ConvSep --
#
#      Converts the separator name (word) into separator character
#
# Argument:
#      separ     Separator word. Is actually called as exp(separ).
#
proc ConvSep {separ} {
global labels
    switch -regexp -- $separ {
        "Tab"     {set sep "\t";  set ext ".tsv"} 
        "Comma"   {set sep ",";   set ext ".csv"}
        "Newline" {switch -- $labels(top) {
	         ""      {set sep "\n";  set ext ".textmerg"} 
	         "%\n"   {set sep "\n";  set ext ".bib"} 
	         default {set sep "\n"; set ext ".tex"}
                 }
	}
        "Blank"   {set sep " " ;  set ext ".dat"}
        "None"    {set sep ""  ;  set ext ".dat"}
        "LaTeX"   {set sep " & "; set ext ".ltx"}
        "LyX"   {set sep "\n\\newline "; set ext ".lyx"}
    }
    return [list $sep $ext]
}

proc ExpNameSet {ext} {
global exportname filename
    if {$exportname == ""} {
        if ![regsub -- ".tk#$" $filename $ext exportname] { 
            set exportname $filename$ext 
        }
    }
}

# Export --
#
#      Export routine, allowing simple data tables with tab, space, comma
#      or newline separators to be written, or LaTeX format table.
#      Newline separated tables are trimmed of whitespaces; first el.
#      will be no. of columns, see procedure Import.
#
proc Export {} {

global initdir exp exportname printcommand
global v headrows rows sidecols cols cw rowsel colsel
global labels
    .w config -cursor watch

    foreach {sep ext} [ConvSep $exp(separ)] {break}
    switch -- $exp(align) {
        {Align Left}    {set al 1}
        {Align Right}   {set al 4}
        {Align Centre}  {set al 8}
    }

    switch -regexp -- $exp(howmuch) {
        All|Col {
                   if {$exp(titlerow)} {
                       set prows [concat $headrows $rows]
                   } else {
                       set prows [ConcatNoTitle $headrows $rows]
                   }
               }
        Selection|Row {
                   set prows $rowsel
                   if {([lsearch $rowsel title] == -1) && $exp(titlerow)} {
                       set prows [concat title $rowsel]
                   }              
               }
    } 

    switch -regexp -- $exp(howmuch) {
        All|Row {
                   if {$exp(titlecol)} {
                       set pcols [concat $sidecols $cols]
                   } else {
                       set pcols [ConcatNoTitle $sidecols $cols]
                   }
               }
        Selection|Col {
                   set pcols $colsel
                   if {([lsearch $colsel title] == -1) && $exp(titlecol)} {
                       set pcols [concat title $colsel]
                   }
               }
    } 

    set typelist [list [list "$ext files" "$ext"] {"All files" {*}}] 


    # add extension, if not yet there, and choose name stem of $filename:
    ExpNameSet $ext

    if ![file isdirectory $initdir] {set initdir [pwd]}

    set exportname [tk_getSaveFile -defaultextension $ext -initialdir $initdir \
         -initialfile "$exportname" -filetypes $typelist -title "tk# file export"]
    
    # Extract piped command 
    set pipepos [string first "|" $exportname]
    incr pipepos 1
    if {$pipepos > 0} {
        set dotpos  [string last  "." $exportname]
        incr dotpos -1
        set printcommand [string range $exportname $pipepos $dotpos]
	set exportname "|$printcommand"
    } 

    if {$exportname != ""} {
	set FileId [open $exportname w] 

        switch -regexp -- $exp(separ) {
            LaTeX {
                puts -nonewline $FileId "\\begin\{tabular\}\{|"
                foreach y $pcols {puts -nonewline $FileId " r |"}
                puts $FileId "\}"
                foreach x $prows {
                    puts $FileId {\hline}
                    foreach y $pcols {
                        if {[lsearch $pcols $y] > 0} {
                            puts -nonewline $FileId $sep
                        } 
                        if ![info exists v($x,$y)] {set v($x,$y) "?6UNDEF"}
                        puts -nonewline $FileId \
                                [string trim [FormattedCell $x $y $cw($y)]] 
                    }
                    puts $FileId "\\\\"
                }
                puts $FileId {\hline}
                puts $FileId "\\end\{tabular\}"
            }
            LyX10|LyX12 {
                if {$exp(separ) == "LyX10"} {
                     # LyX Format 2.10
                     puts $FileId "# This file was created by tk#"
                     puts $FileId "\\lyxformat 2.10"
                     puts $FileId "\\layout Standard"
                     puts $FileId "..."
                     puts $FileId "\\layout Standard"
                     puts $FileId "\\LyXTable"
                     puts $FileId "multicol2"
                     puts $FileId "[llength $prows] [llength $pcols]"
                     foreach x $prows {
                         if {[lsearch $prows $x] != [Pred [llength $prows]]} {
                             puts $FileId "1 0"
                         } else {
                             puts $FileId "1 1"
                         }
                     }
                     foreach x $pcols {
                         if {[lsearch $pcols $x] != [Pred [llength $pcols]]} {
                             puts $FileId "$al 1 0"
                         } else {
                             puts $FileId "$al 1 1"
                         }
                     }
                     foreach x $prows {
                         foreach y $pcols {
                             puts $FileId "0 8 1 0"
                         }
                     }

                } else { 
                     # LyX Format 2.15
                     puts $FileId "# This file was created by tk#"
                     puts $FileId "\\lyxformat 2.15"
                     puts $FileId "\\layout Standard"
                     puts $FileId "..."
                     puts $FileId "\\layout Standard"
                     puts $FileId "\\LyXTable"
                     puts $FileId "multicol4"
                     puts $FileId "[llength $prows] [llength $pcols]\
                             0 0 -1 -1 -1 -1"
                     foreach x $prows {
                         if {[lsearch $prows $x] != [Pred [llength $prows]]} {
                             puts $FileId "1 0 0 0"
                         } else {
                             puts $FileId "1 1 0 0"
                         }
                     }
                     foreach x $pcols {
                         if {[lsearch $pcols $x] != [Pred [llength $pcols]]} {
                             puts $FileId "$al 1 0 "
                         } else {
                             puts $FileId "$al 1 1 "
                         }
                     }
                     foreach x $prows {
                         foreach y $pcols {
                             puts $FileId "0 8 1 0 0 0 0"
                         }
                     }
                puts $FileId ""
                }

                foreach x $prows {
                    foreach y $pcols {
                        if ![info exists v($x,$y)] {
                            set z "?7UNDEF"
                        } else {set z [string trim \
                            [FormattedCell $x $y $cw($y)]]}

                        if {([lsearch $pcols $y] > 0)\
                                || ([lsearch $prows $x] > 0)} {
                            puts $FileId $sep
                        }
                        # Fix backslash for LyX:                         
                        if {[regexp {\\} $z]} {
                            puts -nonewline $FileId {\latex latex }
                            regsub -all {\\} $z {\\backslash } z
                        }
                        puts -nonewline $FileId $z
                        if {[regexp {\\} $z]} {
                            puts $FileId {\latex default}
                        }
                    }
                }
                puts $FileId ""
                puts $FileId "\\layout Standard"
                puts $FileId ""
                puts $FileId "..."
                if {$exp(separ) == "LyX12"} {
                    puts $FileId "\\the_end"
                }
            }
            default {
		if {($exp(separ) == "Newline")} {
                    # Save no. of cols for import:
                    set v(title,title) [llength $pcols]
                    # Only for label printing:
		    puts -nonewline $FileId $labels(top)
		}
                foreach x $prows {
                    if {($exp(separ) == "Newline")} {
                        # BibTeX: split label
			if {$labels(top) == "%\n"} {
                            regsub -all ":" $x "\{" xx
			    set labels(start) "@$xx,\n" 
			}
			puts -nonewline $FileId $labels(start) 
	            }
                    foreach y $pcols {
                        if {[lsearch $pcols $y] > 0} {
                            puts -nonewline $FileId $sep
                        } 
                        # Remove separator char from string:
                        regsub -all $sep [FormattedCell $x $y $cw($y)] "_" z
                        if {$exp(separ) == "Blank"} {
                            regsub -all "_" $z " " z
# gnuplot stuff
#                            set zz 0
#                            scan $z "%g" zz
#                            if {$zz == 0} {
#                                set z [format "%$cw($y)s" 0]
#                            }
                        }
                        if {($exp(separ) == "Newline")} {
                            set z [string trim $z]
			}
                        if {$exp(separ) == "Comma"} {
                            set z \"[string trim $z]\"
                            regsub -all "_" $z " " z
			}
                        if {$labels(top) == "%\n"} {set labels(itemstart) "$y = \{" }
			if {($exp(separ) == "Newline")} {
			    puts -nonewline $FileId $labels(itemstart) 
                        }
			puts -nonewline $FileId $z
                        if {($exp(separ) == "Newline")} {
			    puts -nonewline $FileId $labels(itemend) 
			} 
                    }
                    if {($exp(separ) == "Newline")} {
		        puts -nonewline $FileId $labels(end) 
		    } else {
		    puts $FileId ""
		    }
		}
	    }
        }
        puts -nonewline $FileId $labels(bottom)
        close $FileId
    }

    .w config -cursor {}
    return
}

# BetterSplit --
#
#      Avoid the generation of empty elements due to multiple separator chars
#
proc BetterSplit {line ch} {
    regsub  "$ch$ch" $line "$ch" line
    regsub  "$ch$ch" $line "$ch" line
    regsub  "$ch$ch" $line "$ch" line
    return [split $line "$ch"]
}

# TrackColWidth --
#
#      keep track of widest cell in column
#
# Arguments:
#      cc   Column name
#      z    One cell content string
#
proc TrackColWidth {cc z} {
global cw
    if [info exists cw($cc)] {
        set sl [string length $z]
        if {$sl > $cw($cc)} {
            set cw($cc) $sl 
        }
    } else {set cw($cc) 12}
    return
}

# Import --
#
#      Counterpart of Export. 
#      Tab, comma or blank separated tables WITH headers can be imported;
#      Newline separated tables WITH headers can be imported. 
#              First element MUST be total no. of columns including col
#              of row titles.
#      LaTeX can not be re-imported.
#
proc Import {} {
global initdir exportname exp reload
global f headrows rows sidecols cols cw rowsel colsel
    # 
    .w config -cursor watch

    foreach {sep ext} [ConvSep $exp(separ)] {break}

    set typelist [list [list "$ext files" "$ext"] {"All files" {*}}] 
    
    if ![file isdirectory $initdir] {set initdir [pwd]}
    set exportname [tk_getOpenFile -defaultextension $ext -initialdir $initdir \
     -initialfile $exportname -filetypes $typelist -title "tk# file import"]
    if {($exportname != "") && [file exists $exportname]} {
        set FileId [open $exportname r] 

        if {($exp(separ) == "Newline")} {
            gets $FileId ncols
            set pcols {title}
            for {set ii 1} {$ii < $ncols} {incr ii} { 
                gets $FileId x
                set pcols [concat $pcols [string trim $x " -"]]
            }
	    set prows {}
            while {![eof $FileId]} {
                gets $FileId y
                set rr [string trim $y " -"]
		set prows [concat $prows $rr]
                set iz 0
                foreach cc [lrange $pcols 1 end] {
                    incr iz
                    gets $FileId z 
                    TrackColWidth $cc $z
                    set f($rr,$cc) $z
                }
            }
        } else {
            set prows {}; set pcols {}
            # title row:
            gets $FileId line
            set x [BetterSplit $line $sep]
            # Remove leading "-" and quotes from column names in title row:
            foreach y $x {set pcols [concat $pcols [string trim $y "\"-"]]}
            set iy 0
            while {![eof $FileId]} {
                incr iy
		gets $FileId line
                # works with tab:
                set linelist [BetterSplit $line $sep]
                if {$exp(titlecol)} {
                # Get row name, removing trailing "-":
                   set rr [string trim [lindex $linelist 0] "\"-"]
                } else {
                # Generate row name if not included:
		   set rr r$iy
		}
		set prows [concat $prows $rr]
                # element-in-row counter:
                if {$exp(titlecol)} {set iz 1} else {set iz 0}
		# Added 0 MV 150899:
		foreach z [lrange $linelist $iz end] {
		    set cc [lindex $pcols $iz]
                    incr iz
                    TrackColWidth $cc $z                    
		    # added MV 150899:
		    set f($rr,$cc) [string trim $z "\""] 
		} 
            }
        }
        close $FileId
        
        # Add the missing rows/cols
	foreach x $prows {
	    if {([lsearch $headrows $x] == -1) && ([lsearch $rows $x] == -1)} {
                set rows [concat $rows $x]
            }
        }
        foreach x $pcols {
            if {([lsearch $sidecols $x] == -1) && ([lsearch $cols $x] == -1)} {
                set cols [concat $cols $x]
            }
        }
    }
    .w config -cursor {}
    ReCalc
    if $reload Modified else {set reload 1}
    return
}

# RecalcCell --
#
#      Recalculates a single cell value.
#      Puts ?UNDEF for undefined formulas. 
#      Uses uplevel to globalize any var definitions put by the user 
#      into $f, so they will be available in other cells too
#
proc ReCalcCell {} {
global rows cols cw r c
global v f
    if ![info exists f($r,$c)] {set f($r,$c) "?1UNDEF"}
    uplevel #0 {
        set w $cw($c)
        if [catch {set v($r,$c) [subst $f($r,$c)]} result] {
            if [regexp {\?ERR 4} $result] {
                set v($r,$c) "?ERR 4 RECURSION"
            } else {
                set v($r,$c) "?ERR 4 $result"
            }
        }
    }
    return
}

# WriteFormula --
#
#      After entry of a formula into the entrybox, evaluation and
#      writing back of value result into spreadsheet.
#
proc WriteFormula {} {
global r c f v autocalc rowsel colsel thisformula frq rows cols
    #
    # "thisformula" is the current content of the Formula: box.
    #
    # Temporary save of $r, $c, to allow their use in loops
    # (and availability of loop-$r and loop-$c to user):

    # $rowsel test, so only title rename if intentional:
    if {($c == "title") && ($r != "title") && ([llength $rowsel] == 1)} {
        RowRename ""
    } else {
        # $colsel test ditto:
        if {($r == "title") && ($c != "title")\
                && ([llength $colsel] == 1)} {
            ColRename ""
        } else {
            .w config -cursor watch
            $frq.flag.flag configure -text "Fill..."
            update idletasks
            
            set oldr $r; set oldc $c
            foreach r $rowsel {
                foreach c $colsel {
                    if {$f($r,$c) != $thisformula} {
                        set f($r,$c) $thisformula
                        # Mark modified only if actual formula change:
                        Modified
                    }
                }
            }
            set r $oldr; set c $oldc

            # Is this a BIG selection?
            set bigr [expr ((10 * [llength $rowsel]) / ([llength $rows] + 1))]
            set bigc [expr ((10 * [llength $colsel]) / ([llength $cols] + 1))]
            set big [expr ($bigr * $bigc)]
            # Only update the selected cells if autocalc off:
            # (Remember ReCalc does a ReDisplay)
            if {$autocalc || ($big > 50)} {ReCalc} else {ReCalc 1}
            PaintSelect
        }
    }
    # Prepare for arrow keys use:
    NavFocus 
    PutInfo
    return
}

# PanelRow --
#
#      Returns the row list identity ($rows or $headrows) belonging to this
#      spreadsheet panel.
#
proc PanelRow {panel} {
global rows headrows
    if {($panel == "main") || ($panel == "side")} { 
        return $rows 
    } else {
        return $headrows 
    }
}

# PanelCol --
#
#      Returns the col list identity ($cols or $sidecols) belonging to this
#      spreadsheet panel.
#
proc PanelCol {panel} {
global cols sidecols
    if {($panel == "main") || ($panel == "head")} { 
        return $cols 
    } else {
        return $sidecols 
    }
}

# SetUpForEdit --
#
#      Makes selection equal to current cell and repaint.
#      Set focus to current cell editing window (Formula:)
#
proc SetUpForEdit {} {
global c r rowsel colsel frq
    if {($c != "title")} {set colsel $c}
    if {($r != "title")} {set rowsel $r}
    PaintSelect
    # Set up for data/formula entry:
    EditFocus
    return
}

proc RollFormulaStackUp {} {
global thisformula formulastack
    set aux $formulastack(5)
    foreach i {4 3 2 1} {
        set formulastack([expr ($i + 1)]) $formulastack($i)
    }
    set formulastack(1) $thisformula 
    set thisformula $aux
    return
}

proc RollFormulaStackDown {} {
global thisformula formulastack
    set aux $formulastack(1)
    foreach i {1 2 3 4} {
        set formulastack($i) $formulastack([expr ($i + 1)]) 
    }
    set formulastack(5) $thisformula 
    set thisformula $aux
    return
}

# SetUpEntry --
#
#      Move current cell formula to Formula: -box.
#
proc SetUpEntry {} {
global r c thisformula f cw cwidth
    # Make width entry box current:
    set cwidth $cw($c)
    if ![info exists f($r,$c)] {set f($r,$c) "?2UNDEF"}
    # Get current cell formula for editing:
    RollFormulaStackUp
    set thisformula $f($r,$c)
    return
}

proc ReadFormula {} {
   SetUpEntry
   SetUpForEdit
}

# GetMouse1 --
#
#      Callback routine for mouse button 1 (data entry) in panels.
#      Call GetCell routine to get current row $r and col $c.
#
# Arguments:
#      panel    Panel identity (corner, head, side or main)
#      sw       Switch to avoid reset of selection range and
#               entry of current cell into Formula: box
#
proc GetMouse1 {panel sw} {
global r c rowsel colsel thisformula f cw cwidth paintallowed
    if {$paintallowed == 1} {
        UnPaintCurrentCell
        # Get the proper name of row/column list for this panel:
        set prows [PanelRow $panel]
        set pcols [PanelCol $panel]
        foreach {r c} [GetCell $panel $prows $pcols] {break} 
        if {$sw == 0} SetUpEntry
        SetUpForEdit
        if {$sw == 1} {
            PaintSelect
        }
    }
    if {$paintallowed == -1} {
        set paintallowed 1
        PaintCurrentCell
    }
}

# GetMouse2 --
#
#      Handle mouse button 2 clicks: Mark insert point.
#
# Note: There is an overloading of this mouse button. It is also used
#       to drag the panels. I don't know if this is OK or not (mv).
#
proc GetMouse2 {panel} {
global rows cols headrows sidecols inspoint
    # remove previous insert marks in row or col headers:
    PaintCell "" "" "" noinsert
    set prows [PanelRow $panel]
    set pcols [PanelCol $panel]
    foreach {x y} [GetCell $panel $prows $pcols] {break} 
    if {($x == "title") || ($y == "title")} { 
        PaintCell $panel $x $y insert 
        if {$x != "title"} {
            set inspoint $x
        } elseif {$y != "title"} {
            set inspoint $y
        } else {set inspoint "title" }
    } else {set inspoint {^$} }
}

# GetMouse3 --
#
#      Handle mouse button 3 clicks: Selection bottom right corner.
#
proc GetMouse3 {panel} {
global r c rowsel colsel
global rows cols headrows sidecols
    set prows [PanelRow $panel]
    set pcols [PanelCol $panel]
    foreach {rr cc} [GetCell $panel $prows $pcols] {break}
    set allrows [concat $headrows $rows]
    set i [lsearch $allrows $r]
    # do not include "title" in selection:
    if {($r == "title")} {incr i}
    set j [lsearch $allrows $rr]
    if {($r != "title")} {set rowsel [lrange $allrows $i $j]}
    set allcols [concat $sidecols $cols]
    set i [lsearch $allcols $c]
    if {($c == "title")} {incr i}
    set j [lsearch $allcols $cc]
    if {($c != "title")} {set colsel [lrange $allcols $i $j]}
    PaintSelect 
    return
}

# GetCell --
#
#      Get cell address (row, col names as function value list) 
#      from mouse click.
#
# Arguments:
#      rect      current panel of spreadsheet
#      prows     current row list
#      pcols     current column list
#
proc GetCell {rect prows pcols} {
global cw cs f frq

    # row address:
    set rn -1 
    while {[$frq.$rect compare current >= "1.0 + $rn lines"]} {
        incr rn
    }
    incr rn -1
    set r [lindex $prows $rn]
    if {$r == ""} {set r title} ;# play safe

    # col address (if possible):
    for {set jj 0} {$jj < [llength $pcols]} {incr jj} {
        set cc [lindex $pcols $jj]
        if {[$frq.$rect compare \
                 "current linestart + $cs($cc) chars" <= current]} {
            set c $cc
        } 
    }

    if {(![info exists c]) || (![info exists r])} {set r title; set c title}
    # Make a list containing two names, a row name and a column name, to return.
    return [list $r $c]
}

# RowColPanel --
#
#      Get panel identity as a name string, given $r and $c.
#
proc RowColPanel {r c} {
global headrows sidecols
    if {[lsearch $headrows $r] > -1} {
        set x [expr {([lsearch $sidecols $c] > -1) ? "corner" : "head" }]
    } else {
        set x [expr {([lsearch $sidecols $c] > -1) ? "side" : "main" }]
    }
    return $x
}

proc PanelEndCodeDo {r c panel rr cc} {
     set y 0    
     if {([lsearch $cc $c] == ([llength $cc] - 1))\
             && ([lsearch $rr $r] < ([llength $rr] - 1))} {
         set y 1
     }
     if {([lsearch $rr $r] == ([llength $rr] - 1))\
             && ([lsearch $cc $c] == 0) && ([llength $rr] != 1)}  {
         set y 2
     }
     return $y
}

# PanelEndCode
#
#       Returns the code 0, 1 or 2, depending on the position of the cell
#       with respect to the panel's lower and right/left edge.
#
# Arguments:
#
#       $r $c    row, column
#       $panel   panel name
#
# Returns:
#       Value 1 if cell is last in a row, but not on last row (i.e. 
#               if newline handling on, a \n should be appended).
#       Value 2 if cell is on last row, first cell (i.e. if 
#               newline handling on, a \n should be prepended.
#       Value 0 otherwise.
#
#                     0 0 0 1    
#                     0 0 0 1    
#                     0 0 0 1   
#                     2 0 0 0
#       
proc PanelEndCode {r c panel} {
global headrows sidecols rows cols
    set y 0
    switch $panel {
      corner {set y [PanelEndCodeDo $r $c "corner" $headrows $sidecols]}
      head   {set y [PanelEndCodeDo $r $c "head" $headrows $cols]}
      side   {set y [PanelEndCodeDo $r $c "side" $rows $sidecols]}
      main   {set y [PanelEndCodeDo $r $c "main" $rows $cols]}
    }
    return $y
}

# PaintSelect --
#
#      Paint the selected rows/cols in the titles, and some more (clock cell,
#      current cell).
#
proc PaintSelect {} {
global rows cols headrows sidecols r c tr tc rowsel colsel frq
    $frq.flag.flag configure -text "Paint..."
    update idletasks
    PaintCell "" "" "" notitle

    # add "title" tag only to the selected cells' rows/cols:"
    foreach x $rowsel {
        PaintCell [RowColPanel $x title] $x title title
    }
    foreach y $colsel {
        PaintCell [RowColPanel title $y] title $y title
    }

    # tag the clock cell:
    PaintCell "" "" "" notime
    if [info exists tr] {
        PaintCell [RowColPanel $tr $tc] $tr $tc time
    }
    # tag the current cell:
    PaintCurrentCell

    NavFocus
    return
}

# FormattedCell --
#
#      Returns the string $v($rr,$cc) clipped to column width.
#      Handles undefined v element (may happen!)
#
proc FormattedCell {rr cc w} {
global v
    # Prevent crash 
    if ![info exists v($rr,$cc)] {set v($rr,$cc) "?3UNDEF"}
    return [format "%${w}s" [string range "$v($rr,$cc)" 0 [expr $w - 1]]]
}

# ReDisplayCell --
#
#       Redisplay a single cell in the right attributes
#
# Arguments:
#       $r $c   used as display location.
#       $sw     Switch: insert     insert cell
#                       replace    replace cell (default)
#                       insNL      insert with newline handling acc. to PanelEndCode,
#                       delrow     delete row
#                       delete     delete cell only.
#
proc ReDisplayCell {r c {sw "replace"}} {
global cw v frq
    set w $cw($c)
    set panel [RowColPanel $r $c]
    set x [FormattedCell $r $c $w]
    if {$sw == "insNL"} {
        set ss [PanelEndCode $r $c $panel]
        switch -- $ss {
        1 {set x "$x\n"}
        2 {set x "\n$x"}
        default {}
        }
    }
    foreach {cb ce} [GetCellEnds $r $c] {break}
    PanelsState normal
    switch -regexp $sw {
        delrow  {
            $frq.$panel delete "$ce linestart" "$cb lineend + 1 char"
        }
        {replace|delete} {
             if {[$frq.$panel compare $ce < end]} {
                     $frq.$panel delete $cb $ce
             }
        }
        default {}
    }

    if {[regexp {replace|insNL|insert} $sw]} {
        if {($r == "title") || ($c == "title")} {
            $frq.$panel insert $cb $x underline
        } else {
            $frq.$panel insert $cb $x
        }
    }
    PanelsState disabled
    return
}

# GetCellEnds --
#
#      Return begin- and endpoints (character positions in row) 
#      of a cell in its text widget (panel).
#
# Arguments:
#      r, c   row, col names
#
proc GetCellEnds {r c} {
global rs cs cw headrows rows sidecols cols
    if ![info exists rs($r)] {
        InitIndexRow $headrows
        InitIndexRow $rows
    }
    if ![info exists cs($c)] {
        InitIndexCol $sidecols
        InitIndexCol $cols
    }
    if [info exists rs($r)] {
        set linebgn  $rs($r)
        incr linebgn
        set cellbgn  $cs($c)
        set cellsize $cw($c)
        return [list "$linebgn.$cellbgn" "$linebgn.$cellbgn + $cellsize chars"]
    } else {
        return {1.0 1.0}
    }
}

# PaintCell --
#
#      Put proper attributes on a single cell
#
# Arguments:
#      thispanel    panel to add attributes (tags) to
#      r c          row and col number (LOCAL VARS!)
#      switch       specifies kind of tag to use.
#
proc PaintCell {thispanel r c switch } {
global cw cs rs frq
    foreach {cb ce} [GetCellEnds $r $c] {break}

    PanelsState normal
    foreach panel {main head side corner} { 
        switch -- $switch {
            noinsert   {$frq.$panel tag remove instag   1.0 end} 
            nocurrent  {$frq.$panel tag remove current  1.0 end}
            notime     {$frq.$panel tag remove time     1.0 end}
            notitle    {$frq.$panel tag remove title    1.0 end} 
            nocurtitle {$frq.$panel tag remove curtitle 1.0 end}
            default    {}
        }
    }
    switch -- $switch {
        title    {$frq.$thispanel tag add title    $cb $ce}
        insert   {$frq.$thispanel tag add instag   $cb $ce}
        curtitle {
            $frq.$thispanel tag add curtitle $cb $ce
            $frq.$thispanel see curtitle.last 
            $frq.$thispanel see curtitle.first
        }
        time     {$frq.$thispanel tag add time     $cb $ce}
        current  {
            $frq.$thispanel tag add current  $cb $ce 
            $frq.$thispanel see current.last
            $frq.$thispanel see current.first
        }
    } 
    PanelsState disabled
    return
}

proc ClearAll {} { 
global frq
    PanelsState normal
    foreach panel {main side head corner} {$frq.$panel delete 1.0 end}
    PanelsState disabled
    return
}

# ReCalc --
#
# Recalc everything (sw = 0) or selection (sw = 1)
#
proc ReCalc {{sw 0}} {
global f v rows cols headrows sidecols r c frq rowsel colsel
    .w config -cursor watch

    if {$sw == 0} {
        $frq.flag.flag configure -text "CALC..."
        set prows [ConcatNoTitle $rows $headrows]
        set pcols [ConcatNoTitle $cols $sidecols]
    } else {
         $frq.flag.flag configure -text "Calc..."
        set prows $rowsel
        set pcols $colsel
    }
    update idletasks
    InitIndex
    # r and c must be global :-( for cell evaluation in #0 scope;
    # however r and c must be available at eval time for each cell
    # within the two loops.
    set oldr $r ; set oldc $c
 
    foreach r $prows {
        foreach c $pcols {
             ReCalcCell
        }
    }

    if {$sw == 0} {
        # restore
        set r $oldr ; set c $oldc
        ReDisplay
    } else {
        $frq.flag.flag configure -text "Display...."
        update idletasks
        foreach r $prows {
            foreach c $pcols {
                ReDisplayCell $r $c
            }
        }
        set r $oldr ; set c $oldc
        PaintSelect;
    }


    .w config -cursor {}
    NavFocus    
    return
}

#  ReDisplay --
#
#      Redisplay the whole assembly of four data panels.
#      The three "title panels" are scaled to contain the text.
#      Necessary attributes (tags) are added to the appropriate cells.
#      Also newline characters (\n) added at the right places.
#      For rows and columns, their names are used instead of 
#      evaluated formulas.
#
proc ReDisplay {} {
global f v cw rows cols headrows sidecols rowsel colsel inspoint frq ht wd
global rs
    .w config -cursor watch
    $frq.flag.flag configure -text "Display..."
    update idletasks
    InitIndex
    ReSizeHeaders
    $frq.main   config -wrap none

    ClearAll
    set prows [concat $headrows $rows]
    set pcols [concat $sidecols $cols]
    foreach r $prows {
        set v($r,title) $r-
    } 
    foreach c $pcols {set v(title,$c) -$c} 
    set v(title,title) "title"

    PanelsState normal
    foreach r $prows {
        update idletasks
        # More efficient:
        set rn [lsearch $headrows $r]
        if {$rn > -1} {
            set panel corner
        } else {
            set panel side
        } 
        if {$r == "title"} {
            foreach c $sidecols {
                set x [FormattedCell "title" $c $cw($c)]
                $frq.$panel insert end $x underline
            }
        } else {
            foreach c $sidecols {
                set x [FormattedCell $r $c $cw($c)]
                if {$c == "title"} {
                    $frq.$panel insert end $x underline
                } else {
                    $frq.$panel insert end $x
                }
            }
        }
        # *Very* fast (thanks Ben Allan!)
##        $frq.$panel insert "$rs($r).end + 1 lines" "\n"
	$frq.$panel insert end "\n"
#        $frq.$panel see end

        if {$rn > -1} {
            set panel head
        } else {
            set panel main
        } 
        if {($r == "title")} {
            foreach c $cols {
                set x [FormattedCell $r $c $cw($c)]
                $frq.$panel insert end $x underline
            }
        } else {
            # inline format cmd, not FormattedCell call, for speed
            foreach c $cols {
                set w $cw($c)
                if ![info exists v($r,$c)] {set v($r,$c) "?4UNDEF"}
                $frq.$panel insert end [format "%${w}s" \
                        [string range "$v($r,$c)" 0 [expr $w - 1]]]
            }
        }
##        $frq.$panel insert "$rs($r).end + 1 lines" "\n"
	$frq.$panel insert end "\n"
#        $frq.$panel see end

    }
    # Remove trailing \n from each panel:
    foreach panel {corner head side main} {
        $frq.$panel delete {end - 1 char} end
    }
   
    PaintSelect 

    if {[info exists inspoint]} {
        # re-paint the inspoint title cell!
        if {([lsearch $prows $inspoint] > -1)} {
            set panel [RowColPanel $inspoint title]
            PaintCell $panel $inspoint title insert
        } elseif {([lsearch $pcols $inspoint] > -1)} {
            set panel [RowColPanel title $inspoint]
            PaintCell $panel title $inspoint insert 
        }
    }
    
    PanelsState disabled
    NavFocus
    return
}

# ColWidth --
#
#      Get column width value from entry box
#
proc ColWidth {} {
global cw colsel rows headrows cols sidecols frq autocalc
    Modified

    $frq.flag.flag configure -text "Display...."
    update idletasks
    foreach y $colsel {
        foreach x [concat $headrows $rows] {
            ReDisplayCell $x $y delete
        }
    }

    # Use scan to guarantee integer return:
    set cwidth 12
    scan [.w.fr.box1.ew get] "%d" cwidth 
    if {($cwidth == 0)} {set cwidth 12}
    foreach x $colsel {
        set cw($x) $cwidth
    }

    InitIndexCol $cols
    InitIndexCol $sidecols

    foreach y $colsel {
        foreach x [concat $headrows $rows] {
            ReDisplayCell $x $y insert
        }
    }
    if $autocalc {ReCalc 1}
    ReSizeHeaders
    PaintSelect
    NavFocus
    return
}

#
# Service routines for joint scrolling of all panels & scrollbars.
# Do not touch unless you know your way around.
#
proc HorizScroll {args} {
global frq
    eval [concat $frq.main xview $args]
    eval [concat $frq.head xview $args]
    return
}

proc VertScroll {args} {
global frq
    eval [concat $frq.main yview $args]
    eval [concat $frq.side yview $args]
    return
}

proc BindDragto { x y args } {foreach w $args {$w scan dragto $x $y}}

proc BindMark { x y args } {foreach w $args {$w scan mark $x $y}}

# ColoursSet -- 
#
#      Make all colours valid on panels and tags after modification 
#      through menu.
#
proc ColoursSet {} {
global colours frq
    foreach panel {main side head corner} {
        $frq.$panel config \
                -foreground $colours(textfg) -background $colours(textbg)
    }
    TagsConfig
    return
}

# Wrapper routine for colour editing
proc ColourReconfig {colour} {
global colours
    set colours($colour) [Colour $colours($colour) $colour]
    ColoursSet
    return
}

# FillListBox --
#
#      obtain rowsel/colsel and put into listbox
#
proc FillListBox {rclist pselName} {
upvar $pselName psel
    .sel.lb delete 0 end

    foreach x $rclist {
        .sel.lb insert end $x
    }

    foreach x $rclist {
        set i [lsearch $rclist $x]
        if {[lsearch $psel $x] > -1} {
            .sel.lb selection set $i
        }
    }
    return
}
# PaintInsAll --
#
#      Tag the insert pointer where-ever it is
#
proc PaintInsAll {} {
global headrows rows sidecols cols inspoint
    PaintCell "" "" "" noinsert
    if [info exists inspoint] {
        PaintInsert $headrows corner 1
        PaintInsert $rows     side   1 
        PaintInsert $sidecols corner 0 
        PaintInsert $cols     head   0 
    }
    return
}
# PaintInsert --
#
#      Tag the insert pointer if in current panel
#      sw indicates row or column pointer
#      For usage see proc PaintInsAll
#
proc PaintInsert {rr panel sw} {
global inspoint
    if {([lsearch $rr $inspoint] > -1)} {
        if $sw {
            PaintCell $panel $inspoint title insert
        } else {
            PaintCell $panel title $inspoint insert
        }
    }
    return
}

# ProcListBox
#
#      Obtain cursor selection, put into rowsel/colsel/inspoint, and exit
#
# Arguments:
#      rclist  list of row/column names
#      rc      flag for row select, col select, or insert point choose.
#
proc ProcListBox {rclist rc} {
global rowsel colsel inspoint
global headrows rows sidecols cols
    set ii [.sel.lb curselection]
    set locpsel {}; foreach i $ii {lappend locpsel [lindex $rclist $i]}
    switch -- $rc {
        "r" {set rowsel $locpsel} 
        "c" {set colsel $locpsel}
        "i" {set inspoint [lindex $locpsel 0]}
        "R" {set rowsel [ConcatNoTitle $headrows $rows]}
        "C" {set colsel [ConcatNoTitle $sidecols $cols]}
        "I" {bell}
    }
    .sel.lb selection clear 0 end
    wm withdraw .sel
    PaintSelect
    PaintInsAll

    return
}

# GetSel --
#
#      Get selection (rows, columns) by string matching with row/col names
#
# Arguments:
#      rowselect   match string for row names
#      colselect   match string for column names
#
proc GetSel {rowselect colselect} {
global rowsel colsel rows headrows cols sidecols
    set rowsel {}
    foreach r [concat $rows $headrows] {
        if {[regexp -- $rowselect $r]} {
            if {$r != "title"} {
                lappend rowsel $r
            }
        }
    }
    set colsel {}
    foreach c [concat $cols $sidecols] {
        if {[regexp -- $colselect $c]} {
            if {$c != "title"} {
                lappend colsel $c
            }
        }
    }
    PaintSelect
    .w config -cursor {}
    wm withdraw .selm
    return
}

proc AllButtonState {rc} {
    if {$rc == "i"} {
        .sel.f1.all config -state disabled
    } else {
        .sel.f1.all config -state normal
    }
}

# SelectEdit --
#
#      Listbox user interface handling routine.
#      Handles the graphics part of the job
#
# Arguments:
#      rclist   list of rownames, col names, or both (to select from)
#      rc       is this row (r), column (c) or both (i) to select
#               a number of elements (r,c) or one insert point (i) from.
#      title    Goes to the subwindow title
#      mode     Listbox selection mode: multiple (r,c) or single (i).
#      selName  variable (list) to output selection into, typically 
#               rowsel, colsel, inspoint.
#
proc SelectEdit {rclist rc title mode selName} {
upvar $selName sel
    .w config -cursor watch
    set doScript  [list ProcListBox $rclist $rc]
    set allScript [list ProcListBox $rclist [string toupper $rc]]
    if [winfo exists .sel] { 
        wm deiconify .sel 
        wm title .sel $title
        raise .sel 
        .sel.lb config -selectmode $mode 
        FillListBox $rclist sel
        set doScript [list ProcListBox $rclist $rc]
        set allScript [list ProcListBox $rclist [string toupper $rc]]
        .sel.f1.do  config -command $doScript
        .sel.f1.all config -command $allScript
        AllButtonState $rc
        .w config -cursor {}
        return
    }
    toplevel .sel
    wm title .sel $title
    CenterWin .sel .w

    foreach x {0 1} {frame .sel.f$x}
    listbox   .sel.lb -selectmode $mode \
            -yscrollcommand {.sel.sb set} -width 30 
    scrollbar .sel.sb -command {.sel.lb yview} -width 15
    FillListBox $rclist sel

    button .sel.f1.do -text OK   -command $doScript
    bind   .sel <Key-Return>              $doScript
    button .sel.f1.all -text All -command $allScript
    AllButtonState $rc
    button .sel.f1.cancel -text cancel -command {wm withdraw .sel} 
    bind   .sel <Key-Escape>                    {wm withdraw .sel}
    
    pack .sel.f1.do     -side left
    pack .sel.f1.all   -side left
    pack .sel.f1.cancel -side right

    grid columnconfigure .sel 0 -weight 1
    grid rowconfigure    .sel 0 -weight 1
    grid .sel.lb .sel.sb  -in .sel -sticky news
    grid .sel.f1  -       -in .sel -sticky news

    raise .sel
    PaintSelect
    .w config -cursor {}
    return
}

# SelectMatch --
#
#      Selection by string matching user interface handling routine
#
proc SelectMatch {} {
global colours rowsel colsel headrows rows sidecols cols
    .w config -cursor watch

    if [winfo exists .selm] { 
        wm deiconify .selm; 
        raise .selm; return
        }
    toplevel .selm
    wm title .selm "Row/Col Select by Match"
    CenterWin .selm .w

    foreach x {1 2 3} {frame .selm.f$x}
    
    foreach {x y} {1 Row 2 Col} {
        label .selm.f$x.l$x -text "$y Select String:" -width 18 -anchor e
        entry .selm.f$x.e$x -textvariable ${y}SelString \
                                             -background $colours(textbg)
    }

    button .selm.f3.do -text OK -command {GetSel $RowSelString $ColSelString} 
    bind   .selm <Key-Return>            {GetSel $RowSelString $ColSelString}

    button .selm.f3.cancel -text cancel -command [list DoNothing .selm] 
    bind   .selm <Key-Escape>                    [list DoNothing .selm]
 
    foreach x {1 2} {
        pack .selm.f$x.l$x -side left     
        pack .selm.f$x.e$x -side left -fill x -expand true
    }
    
    pack .selm.f3.do -side left
    pack .selm.f3.cancel -side right
    
    foreach x {1 2 3} {grid .selm.f$x -in .selm -row $x -sticky ew}

    raise .selm
    return
}


# RCSpawn --
#
#      Service routine for row/column spawning.
#
# Arguments:
#      rr1     row/col list 1
#      rr2     row/col list 2
#      switch  r for rows, c for columns
#  
proc RCSpawn {rr1Name rr2Name switch} {
global r c inspoint rows headrows cols sidecols f v cw 
global template spawn
    upvar $rr1Name rr1
    upvar $rr2Name rr2
    set x 1
    scan $spawn "%d" x
    set spawn $x

    for {set in 0} {$in < $spawn} {incr in} {

        set i "($in)"
        # template may contain $i!
        set t [subst $template]
        if ![NewNameError [concat $rr1 $rr2] $t] {
            set ii [lsearch $rr1 $inspoint]
            if {$ii > -1} {
                # insert AFTER first element:
                set rr1 [concat [lrange $rr1 0 [Pred $ii]] $t \
                                             [lrange $rr1 [This $ii] end]] 
            } else {
                set ii [lsearch $rr2 $inspoint]
                if {$ii > -1} {
                    # insert BEFORE 1st el:
                    set rr2 [concat [lrange $rr2 0 [expr $ii - 1]] $t \
                                             [lrange $rr2 $ii end]] 
                } else { 
                    set rr2 [concat $rr2 $t]
                }
            }
            # copy over f,v,cw values:
            if {$switch == "r"} {
                set f($t,title) "-"
                set v($t,title) "$t-"
                ReDisplayCell $t "title" "insNL"
                set ylist [ConcatNoTitle $sidecols $cols]
                foreach y $ylist {
                    if [info exists f($r,$y)] {set f($t,$y) $f($r,$y)}
                    if [info exists v($r,$y)] {set v($t,$y) $v($r,$y)}
                    ReDisplayCell $t $y "insNL"
                } 

            } else {
                set f(title,$t) "-"
                set v(title,$t) "-$t"
                set cw($t) $cw($c)
                ReDisplayCell "title" $t "insert"
                set xlist [ConcatNoTitle $headrows $rows]
                foreach x $xlist {
                    if [info exists f($x,$c)] {set f($x,$t) $f($x,$c)}
                    if [info exists v($x,$c)] {set v($x,$t) $v($x,$c)}
                    ReDisplayCell $x $t "insert"
                }
            }
        } 
        update idletasks
    }
    PaintSelect
    Modified
    TemplateDefault
    .w config -cursor {}
    return
}

# RowSpawn --
#
#      Wrapper routine for row(s) spawn. Mainly uses RCSpawn.
proc RowSpawn {win} {
global rows headrows autocalc inspoint
    RCSpawn headrows rows r 
    if $autocalc ReCalc else ReSizeHeaders
    wm withdraw $win
    return
}

# AddRow --
#
#      Add a row to bottom of spreadsheet. Name is generated from no.
#      of rows, with added robustness :-)
#
proc AddRow {} {
global spawn template inspoint r c rows headrows cols sidecols 
global autocalc defmov thisformula f v frq
    set oldinspoint $inspoint
    set inspoint {^$}
    set spawn 1
    set ii [llength $rows]
    if {$thisformula == ""} {
        set template "R<$ii>"
    } else {
        set template $thisformula
    }
    while {[NewNameError [concat $headrows $rows] $template]} {
        incr ii
        set template "R<$ii>"
    }
    RCSpawn headrows rows r
    set inspoint $oldinspoint
    set r [lindex $rows [expr [llength $rows] - 1]]
    set c [lindex $cols 0]
    if {$c == ""} {set c title}
    PaintCell "" "" "" nocurrent
    PaintCell [RowColPanel $r $c] $r $c current
    foreach x [ConcatNoTitle $sidecols $cols] {
        set f($r,$x) "-"
        set v($r,$x) "-"
        ReDisplayCell $r $x 
    }
    SetUpForEdit
    set defmov {>}
    $frq.flag.mv configure -text $defmov -font fixed
    if $autocalc ReCalc else ReSizeHeaders
    return
}

# ColSpawn --
#
#      Column(s) spawn wrapper routine. Uses mainly RCSpawn for
#      its functionality.
#
proc ColSpawn {win} {
global cols sidecols autocalc inspoint
    RCSpawn sidecols cols c
    if $autocalc ReCalc else ReSizeHeaders
    wm withdraw $win
    return
}

# ReplaceElement --
#
#      Replaces a single element in a list of row/col names.
#      Used in row /column rename routines.
#
# Arguments:
#      rr      list of elements to work on
#      r       the old, present element
#      newr    the new element (name) to replace it with.
#
proc ReplaceElement {rr r newr} {
    set ii [lsearch $rr $r]
    if {($ii > -1)} {
        return [lreplace $rr $ii $ii $newr]
    } else {
        return $rr
    }     
}

# NewNameError --
#
#      Report error in rename row/col, such as empty template or 
#      attempt at using already existing name
# 
# Arguments:
#      rr    List of existing row or col names
#      newr  The new name template proposed
#      oldr  (optional) old name which is modified, even if 
#            contains/contained in new one.
#
# Results:
#      Error code 1 in case of error
#
proc NewNameError {rr newr {oldr ""}} {
global template
    # Empty template:
    if {$newr == ""} {
        return 1
    }
    # Contains blanks:
    if {[string first " " $newr] > -1} {return 1}
    # Contains/is contained in already:
    foreach x $rr {
        if {$x == $newr} {
            if {$x != $oldr} {return 1}
        }

#        if {([string first $newr $x] > -1) ||\
#                 ([string first $x $newr] > -1)} {
#            if {$x != $oldr} {return 1}
#        }
    } 
    return 0
}

# RowRename --
#
#      Rename selected row(s). If given name (acc. to template)
#      already exists, don't do it.
#
# Argument:
#      win   The auxiliary window containing the dialogue.
#
proc RowRename {win} {
global rows headrows sidecols cols r c f v rs rowsel colsel thisformula

    set oldr $r
    set oldc $c

    set in 0
    foreach r $rowsel {
        if {$r != "title"} {
            incr in
            set i "($in)"
            if {$win == ""} {
                set template $thisformula
            } else {
                set template [subst [$win.f0.e get]] 
            }
            if ![NewNameError [concat $headrows $rows] $template $r] {
                set headrows [ReplaceElement $headrows $r $template]
                set rows     [ReplaceElement $rows $r $template]
                set rowsel   [ReplaceElement $rowsel $r $template]
                foreach c [concat $sidecols $cols] {
                    if [info exists f($r,$c)] {
                        set f($template,$c) $f($r,$c)
                        unset f($r,$c)
                    } 
                    if [info exists v($r,$c)] {
                        set v($template,$c) $v($r,$c)
                        unset v($r,$c)
                    }
                }
                set v($template,title) ${template}-
                if [info exists rs($r)] {
                    set rs($template) $rs($r); unset rs($r)
                } else {puts "RowRename: rs($r) not exist bug"}
                if {$r == $oldr} {set oldr $template}
            }
        } 
    }
    set r $oldr
    set c $oldc

    # Refresh the actual title cell:
    foreach x $rowsel {ReDisplayCell $x title} 
    PaintSelect

    if {$win != ""} {wm withdraw $win} 
    .w config -cursor {}
    Modified
    return
}

# ColRename --
#
#      Rename selected cols. If given name (acc. to template)
#      already exists, don't do it.
#
# Argument:
#      win   The auxiliary window containing the dialogue.
#
proc ColRename {win} {
global cols sidecols rows headrows r c f v cs cw colsel rowsel thisformula
    set oldr $r
    set oldc $c
    
    set in 0
    foreach c $colsel {
        if {$c != "title"} {
            incr in
            set i "($in)"
            if {$win == ""} {
                set template $thisformula
            } else {
                set template [subst [$win.f0.e get]]
            }
            if ![NewNameError [concat $sidecols $cols] $template $c] {
                set sidecols [ReplaceElement $sidecols $c $template]
                set cols     [ReplaceElement $cols $c $template]
                set colsel   [ReplaceElement $colsel $c $template]
                foreach r [concat $headrows $rows] {
                    if [info exists f($r,$c)] { 
                        set f($r,$template) $f($r,$c) 
                        unset f($r,$c)
                    }
                    if [info exists v($r,$c)] {
                    set v($r,$template) $v($r,$c) 
                    unset v($r,$c)
                    }
                }
                set v(title,$template) -${template}
                if [info exists cs($c)] {
                    set cs($template) $cs($c); unset cs($c)
                } else {puts "ColRename: cs($c) not exist bug"}
                if [info exists cw($c)] {
                    set cw($template) $cw($c); unset cw($c)
                } else {puts "ColRename: cw($c) not exist bug"}
                if {($c == $oldc)} {set oldc $template}
            }
        }
    }

    set r $oldr
    set c $oldc

    # Refresh the actual title cell:
    foreach x $colsel {ReDisplayCell title $x} 
    PaintSelect

    if {$win != ""} {wm withdraw $win}
    .w config -cursor {}
    Modified
    return
}

# SplitSel
#
#      Split rr1, rr2 into "selected" (s) and "rest" (x1, x2)
#
# Arguments:
#      rr1, rr2   row/col name lists
#      rs         row/col selection list
#      s          output selected names
#      x1,x2      output rest
#
proc SplitSel {rr1 rr2 rs} {
    set s {}; set x1 {}; set x2 {}
    foreach y $rr1 {
        if {([lsearch $rs $y] == -1)} {lappend x1 $y} else {lappend s $y}
    }
    foreach y $rr2 {
        if {([lsearch $rs $y] == -1)} {lappend x2 $y} else {lappend s $y}
    }
    return [list $x1 $x2 $s]    
}

# CellAbove ... CellRight --
#
#      Procedures referring to the value in the neighbouring cell
#
proc CellAbove {} {
global headrows rows r c v
    set itemlist [concat $headrows $rows]
    set x [lindex $itemlist [Pred [lsearch $itemlist $r] ] ]
    return $v($x,$c)
}

proc CellBelow {} {
global headrows rows r c v
    set itemlist [concat $headrows $rows]
    set x [lindex $itemlist \
            [Succ [lsearch $itemlist $r] [llength $itemlist] ] ]
    return $v($x,$c)
}

proc CellLeft {} {
global sidecols cols r c v
    set itemlist [concat $sidecols $cols]
    set x [lindex $itemlist [Pred [lsearch $itemlist $c] ] ]
    return $v($r,$x)
}

proc CellRight {} {
global sidecols cols r c v
    set itemlist [concat $sidecols $cols]
    set x [lindex $itemlist \
            [Succ [lsearch $itemlist $c] [llength $itemlist] ] ]
    return $v($r,$x)
}

#
# Functions Pred (predecessor), This, and Succ (successor), supporting
# moves to neighbouring cells.
#
proc Pred {i} {
    set i [expr ($i - 1)]
    if {$i < 0} {set i 0}
    return $i
}

proc This {i} {
    if {$i < 1} {set i 1}
    return $i
}

proc Succ {i max} {
    incr i
    if {$i >= $max} {set i 0}
    return $i
}

# RCMove --
#
#      Service routine for row/col move.
#
# Arguments:
#      rr1,rr2    Row or col name pair, like $headrows,$rows
#      rs         Corresponding selection range, like $rowsel
#      sw         Optional switch for sort selection
#
# Result:
#      rs is extracted (and removed) from [concat $rr1 $rr2] and 
#      either inserted before element $inspoint, or appended to $rr2.
#
proc RCMove {rr1 rr2 rs} {
global inspoint 

    foreach {x1 x2 s} [SplitSel $rr1 $rr2 $rs] {break}
    
    # ii is Insert Index
    set ii [lsearch $x1 $inspoint]
    if {$ii > -1} {
        return [list \
         [concat [lrange $x1 0 [Pred $ii]] $s [lrange $x1 [This $ii] end]] \
            $x2]
    } else {
        set ii [lsearch $x2 $inspoint]
        if {$ii > -1} {
            return [list $x1 \
                [concat [lrange $x2 0 [expr ($ii - 1)]] $s [lrange $x2 $ii end]]]
        } else {return [list $x1 [concat $x2 $s]]}
    }
    Modified
}

proc ReSizeHeaders {} {
global headrows sidecols frq ht wd
     set ht [InitIndexRow $headrows]
     set wd [InitIndexCol $sidecols] 
     $frq.side   config -width  $wd 
     $frq.corner config -width  $wd -height $ht 
     $frq.head   config -height $ht 
     return
}

# RowMove, ColMove --
#
#      Moves one or more rows or cols (selection interval)
#      to the insert pointer $inspoint. If not set (appropriately),
#      append to end of worksheet.
#
# Argument:
#      sw    if 1: sort the rows/cols moved.
#
proc RowMove {{sw 0}} {
global headrows rows rowsel sidecols cols inspoint
    .w config -cursor watch
    if $sw {set rowsel [lsort $rowsel]}
    foreach x $rowsel {
        ReDisplayCell $x [lindex $sidecols 0] "delrow"
        if {[llength $cols] > 0} {
            ReDisplayCell $x [lindex $cols 0] "delrow" 
        }
        foreach {headrows rows} [RCMove $headrows $rows $x] {break}
        InitIndexRow $headrows
        InitIndexRow $rows
        foreach y [concat $sidecols $cols] {ReDisplayCell $x $y "insNL"}
    }
    ReSizeHeaders
    PaintSelect
    .w config -cursor {}
    return
}

proc ColMove {{sw 0}} {
global sidecols cols colsel headrows rows inspoint
    .w config -cursor watch
    if $sw {set colsel [lsort $colsel]}
    foreach y $colsel {
        foreach x [concat $headrows $rows] {ReDisplayCell $x $y "delete"}
        foreach {sidecols cols} [RCMove $sidecols $cols $y] {break}
        InitIndexCol $sidecols
        InitIndexCol $cols
        foreach x [concat $headrows $rows] {ReDisplayCell $x $y "insert"}
    }
    ReSizeHeaders
    PaintSelect
    .w config -cursor {}
    return
}

# RCKill --
#
#      Service routine for row/column kill.
#
# Arguments:
#      rrName    list of row/col names to choose from.
#      rs        Name of row/col to kill.
#
proc RCKill {rrName rs} {
    upvar $rrName rr
    set x {}
    foreach y $rr {if {([lsearch $rs $y] == -1)} {lappend x $y}}
    set rr $x
    Modified
    return
}

proc RowKill {} {
global r rowsel rows headrows cols sidecols autocalc frq rs
    foreach x $rowsel {
        ReDisplayCell $x [lindex $sidecols 0] "delrow"
        ReDisplayCell $x [lindex $cols 0]     "delrow"
        RCKill headrows $x
        RCKill rows $x
        InitIndexRow $headrows
        InitIndexRow $rows
    } 
    set r title ;# safe
    set rowsel {}
    if $autocalc ReCalc
    return
}

proc ColKill {} {
global c colsel cols sidecols rows headrows autocalc
    foreach y $colsel {
        foreach x [concat $headrows $rows] {ReDisplayCell $x $y "delete"}
        RCKill sidecols $y
        RCKill cols $y
        InitIndexCol $sidecols
        InitIndexCol $cols
   }
    set colsel {}
    set c title
    if $autocalc ReCalc
    return
}

#
# Executed when regretting a subwindow operation (pressing Cancel)
#
proc DoNothing {win} {
    wm withdraw $win
    .w config -cursor {}
}

# RCSpawnWindow
#
#      Window user interface for row/col spawn data input.
#
# Arguments:
#      title       Goes to title of sub-window.
#      spawnCmd    The command RowSpawn or ColSpawn to be executed.
#      win         Window identifier used. Is used by RowSpawn/ColSpawn
#                  to close this window after job done.
#
proc RCspawnWindow {title spawnCmd win} {
global template spawn colours
    .w config -cursor watch
    if [winfo exists $win] { wm deiconify $win; raise $win; return }

    toplevel $win
    wm title $win $title 
    CenterWin $win .w
    wm resizable $win true false

    frame $win.f1
    frame $win.f2
    frame $win.f3

    label    $win.f1.l1 -text "Spawn Factor:" -width 15 -anchor e
    entry    $win.f1.e1 -textvariable spawn -width 5 -background $colours(textbg)

    label    $win.f2.lfac -text "Template:" -width 15 -anchor e
    entry    $win.f2.efac -textvariable template -background $colours(textbg)

    set com1 [list $spawnCmd $win]
    set com2 [list DoNothing $win]
    button $win.f3.ok    -text OK  -command $com1 
    bind     $win <Key-Return> $com1
    label    $win.f3.info    -text {-$i-} -justify center
    button $win.f3.close -text Cancel -command $com2
    bind     $win <Key-Escape> $com2

    pack $win.f1.l1 -side left 
    pack $win.f1.e1 -side left
    pack $win.f2.lfac -side left
    pack $win.f2.efac -fill x -expand true -side left
    pack $win.f3.ok -side left
    pack $win.f3.info -fill x -expand true -side left
    pack $win.f3.close -side right

    grid columnconf $win 0 -weight 1
    grid $win.f1 -in $win -column 0 -row 0 -sticky ew 
    grid $win.f2 -in $win -column 0 -row 1 -sticky ew 
    grid $win.f3 -in $win -column 0 -row 2 -sticky ew 
    
    raise $win
    return
}

# ShowProgress --
#
#       Show the progress of an operation (num) in specified  
#       entry box (box)
#
proc ShowProgress {box num} {
    $box delete 0 end
    $box insert end $num
    update idletasks
}

proc Geometry {win} {
global headrows rows sidecols cols v f cw frq
    # Make for a good start:
    .w config -cursor watch
    $frq.flag.flag configure -text "Rows..."
    update idletasks

    set ispawn [$win.e1 get]
    set jspawn [$win.e2 get]

    # Limit appending work for large sheets: - Ben Allan, there you go!
    set oldrows $rows
    set temprows {}

    for {set i 0} {$i < $ispawn} {incr i} {
        if {(1000 * ($i / 1000)) == $i} {
            ShowProgress $win.e1 $i
            set rows [concat $rows $temprows]
            set temprows {}
        }
        set r "R<$i>"
        # collect in temporary...
        if {[lsearch $oldrows $r] == -1} {
            set temprows [lappend temprows $r]
        }
    }
    ShowProgress $win.e1 $i
    set rows [concat $rows $temprows]

    $frq.flag.flag configure -text "Fill..."
    update idletasks

    set oldcols $cols

    for {set j 0} {$j < $jspawn} {incr j} {
        set c "C<$j>"
        if {[lsearch $oldcols $c] == -1} {
            set cols [lappend cols $c]
            ShowProgress $win.e2 $j
        }
        if ![info exists cw($c)] {set cw($c) 12}
        for {set i 0} {$i < $ispawn} {incr i} {
            set r "R<$i>"
            if {(1000 * ($i / 1000)) == $i} {
                ShowProgress $win.e1 $i
            }
            if ![info exists f($r,$c)] {set f($r,$c) "-"}
            if ![info exists v($r,$c)] {set v($r,$c) "-"}
        }
        ShowProgress $win.e1 $i
        ShowProgress $win.e2 $j
    }
    ReDisplay
    .w config -cursor {}
    wm withdraw $win
    return
}

proc GetGeometry {win} {
    if [winfo exists $win] { wm deiconify $win; raise $win; return }
    .w config -cursor watch
    toplevel $win
    wm title $win "Spreadsheet Geometry"
    wm resizable $win false false
    CenterWin $win .w

    label  $win.l1 -width 15 -text "No. of Rows:" -anchor w
    label  $win.l2 -width 15 -text "No. of Columns:" -anchor w
    entry  $win.e1 -width 15 
    entry  $win.e2 -width 15
    button $win.do   -text "OK"     -anchor w \
            -command [list Geometry $win]
    bind   $win <Key-Return> [list Geometry $win]
    button $win.dont -text "Cancel" -anchor e \
            -command [list DoNothing $win]
    bind   $win <Key-Escape> [list DoNothing $win]

    grid   $win.l1 $win.l2   -in $win -sticky news
    grid   $win.e1 $win.e2   -in $win -sticky news
    grid   $win.do   -in $win -row 2 -column 0 -sticky w
    grid   $win.dont -in $win -row 2 -column 1 -sticky e

    raise $win
    return
}

# RCrenameWindow --
#
#      User interface service routine for inputting a string, 
#      typically a template string for row/col rename.
#
# Arguments:
#      title      To appear in window title.
#      labeltext  To appear as label by the entry box.
#      cmd        Callback cmd to be associated with OK.
#      win        identifier of new toplevel window to be opened/used.
#                 RowRename/ColRename will finally close this.
#      rc         r for row rename, c for col rename
#
proc RCrenameWindow {title labeltext renameCmd win rc} {
global colours
    if [winfo exists $win] { wm deiconify $win; raise $win; return }
    .w config -cursor watch
    toplevel $win
    wm title $win $title
    CenterWin $win .w
    wm resizable $win true false

    frame    $win.f0
    frame    $win.f1

    label    $win.f0.l -text $labeltext -anchor e
    entry    $win.f0.e -textvariable template -background $colours(textbg)

    set doScript     [list $renameCmd  $win] 
    set cancelScript [list DoNothing   $win]

    button $win.f1.tag   -text "Tag" -command "set template \\\$$rc#"
    button $win.f1.untag -text "Untag" -command \
            "set template {\[string trimright \$$rc \"#\"\]}"
    button $win.f1.do  -text "OK"       -command $doScript
    bind   $win <Key-Return>                     $doScript
    button $win.f1.close -text "Cancel" -command $cancelScript 
    bind   $win <Key-Escape>                     $cancelScript
    label  $win.f1.info  -text {-$i-} -justify center

    pack $win.f0.l -side left
    pack $win.f0.e -fill x -expand true -side left
    pack $win.f1.do $win.f1.tag $win.f1.untag -side left
    pack $win.f1.info -fill x -expand true -side left
    pack $win.f1.close -side right

    grid columnconf $win 0 -weight 1
    grid $win.f0 -in $win -sticky ew 
    grid $win.f1 -in $win -sticky ew 
    
    raise $win
    return
}

# Now, Today, USAtoday --
#
#      Return current time, formatted, for use in clock applications and 
#      user file name generation.
#
proc Now {} {
#    set x [clock format [clock seconds]]
    set x [clock format [clock seconds] -format "%Y.%m.%d=%T"]
#    regsub -all { } $x {=} x
    return $x
}

proc Today {} {
    return [clock format [clock seconds] -format {%d-%h-%Y}]
}

proc USAtoday {} {
    return [clock format [clock seconds] -format {%Y/%h/%d}]
}

# StartClock, StopClock --
#
#      Routines to put a running clock (hours) into a cell, and to stop it.
#      (this could be generalized by using time formatting; reminder mv)
#
proc StartClock {} {
global r c tr tc f thisformula rows cw frq sidecols cols v tools
    if [info exists tr] {StopClock}
    $tools.stop config -state normal
    # Make room for date strings:
    set cw(title) 36
    lappend rows [Now]

    # last (new) row:
    set r [lindex $rows [expr ([llength $rows] - 1)]]

    set fmt [.w.fr.box1.ef get]
    if {$fmt == ""} {set fmt {$cw($c).2}}
    set thisformula "\[Fmt \"%${fmt}f\" \"\"\
     \[expr (\[clock seconds\] - [clock seconds]) / 3600.0 \]\]"
    foreach x [concat $sidecols $cols] {set f($r,$x) "-"; set v($r,$x) "-"}
    set f($r,$c) $thisformula
    set v($r,$c) CLOCK
    set tr $r; set tc $c
    ReDisplay
    Modified
}

# StopClock --
#
#      Stops a running clock if there is one.
#      if an emergency stop, do not try to reconfig
#      any buttons (they have already disappeared)
#  
# Argument:
#      sw    1 if emergency stop
#
proc StopClock {{sw 0}} {
global r c f v thisformula tr tc tools
    if [info exists tr] {
        if [winfo exists .w] {
            $tools.stop config -state disabled
        }
        # Make sure that CLOCK cell is stopped:
        set oldr $r; set oldc $c
        set r $tr; set c $tc
        ReCalcCell
        if [info exists v($r,$c)] {
            set thisformula [string trim $v($r,$c)]
        } else {set thisformula "?5UNDEF"}
        set f($r,$c) $thisformula
        Modified
        if [winfo exists .w] {
            ReDisplayCell $r $c
            PaintCell "" "" "" notime
        }
        unset tr; unset tc
        set r $oldr; set c $oldc
    }
}

# Fmt, Expr --
#
#      Routines to be used in Format box, to implement in an elegant way
#      formatting/unformatting of entries.
#
# Arguments:
#      fmt1   format to be used by [format...] stmt
#      fmt2   format to be used by [scan...] stmt; if "", no scan
#      in     input data (current cell content)
#
proc Fmt {fmt1 fmt2 {in 0} args} {
    if {$fmt2 == ""} {
        return "[format $fmt1 $in]"
    } else {
        return "[format $fmt1 [set x 0; scan $in $fmt2 x; set x]]"
    }
}

proc Expr {{in 0}} {
    catch {set result [expr ($in)]} result 
    return $result
}

# ComposeCmd --
#
#      Service routine for the editing of formatting commands 
# Argument:
#      fmt          Input formatted formula string
#
# Result:
#      thisformula  The variable associated with Formula entrybox
#
proc ComposeCmd {fmt} {
global thisformula
    set thisformula $fmt 
    return
}    

#  ComposeFixed ... ComposeCurr, AddExpr, LeftAlign --
#
#      The formatted string composition routines, used in the 
#      Formula box when editing.
#
proc ComposeFixed {} {
global ef e
    ComposeCmd "\[Fmt \"%[$ef get]f\" \"%g\" [$e get] \]"
    return
}

proc ComposeFloat {} {
global ef e
    ComposeCmd "\[Fmt \"%[$ef get]g\" \"%g\" [$e get] \]"
    return
}

proc ComposeInt {} {
global ef e
    ComposeCmd "\[Fmt \"%[$ef get]d\" \"%d\" [$e get] \]"
    return
}

proc ComposeCurr {} {
global w e
    ComposeCmd "\[Fmt \"%$\{w\}.2f\" \"%g\" [$e get] \]" 
    return
}

proc AddExpr {} {
global e 
    ComposeCmd "\[Expr [$e get]\]"
    return
}

proc LeftAlignText {} {
global w e
    ComposeCmd "\[Fmt \"%-$\{w\}s\" \"\" \{[$e get]\}\]"
    return
}

proc LeftAlign {} {
global w e
    ComposeCmd "\[Fmt \"%-$\{w\}s\" \"%s\" [$e get]\]"
    return
}

proc RemoveFormat {} {
global e
    if [regexp {^\[Fmt +"[^"]*" +"[^"]*" +(.*)\] *$} \
            [$e get] dummy y] {
        ComposeCmd $y
    }
    # Remove excess curlies if necessary (for leftalign multiword):
    if [regexp {^\{(.*)\}} [$e get] dummy y] {
        ComposeCmd $y
    }
}

proc RemoveExpr {} {
global e
    if [regexp {^\[Expr *(.*) *\] *$} [$e get] dummy y] {
        ComposeCmd $y
    }
    return
}

# FormatSelect --
#
#      Service routine allowing the application of a Formula: -box
#      reformatting stmt to the whole current selection.
#
# Argument:
#      formatScript   One of the "Compose..." cmds for reformatting
#                     or "unformatting". 
#
proc FormatSelect {formatScript} {
global rowsel colsel r c thisformula f autocalc
    set oldr $r
    set oldc $c
    foreach r $rowsel {
        foreach c $colsel {
            SetUpEntry
            $formatScript
            set f($r,$c) $thisformula
        }
    }
    set r $oldr
    set c $oldc
    if $autocalc ReCalc else {ReCalc 1}
    return
}

# PrintSel --
#
# Sends the current selection to the printer.
#
proc PrintSel {rows cols} {
global v printcommand thisformula cw

    if {$thisformula != ""} {
	set printcommand $thisformula 
    } else { set thisformula $printcommand }

    set myfile [open "|$printcommand" w]
    
    foreach r $rows {
        foreach c $cols {
            puts -nonewline $myfile [FormattedCell $r $c $cw($c)]
        }
    puts $myfile ""
    }

    # occasionally needed error report:
    #set code [catch {close $myfile} result] 
    #if {$code > 0} { puts "TCL code $code: $result" }

    close $myfile
    return
}

# CenterWin --
#
#      Centers a window (win) in its parent (otherwin). 
#      Improved version of one found in Midnight Commander.
#
proc CenterWin {win otherwin} {
    wm withdraw $win
    foreach {c r} {x height y width} {set cen$c [winfo req$r $win]}
    set geo [split [wm geometry $otherwin] +x]
    foreach {p n} {pdx 0 pdy 1 px 2 py 3} {set $p [lindex $geo $n]}
    foreach c {x y} {set $c [expr \$p$c+((\$pd$c-\$cen$c)/2)]}
    wm geometry $win +$x+$y
    wm deiconify $win
    return
}

# TagsConfig --
#
#      (re-)Configure the colour tags used in spreadsheet panels
#
proc TagsConfig {} {
global frq colours
    foreach panel {main side head corner} {
        $frq.$panel tag config title -background $colours(selcolour)
        $frq.$panel tag config time  -foreground $colours(timecolour) 
        $frq.$panel tag config instag -underline true\
                       -foreground $colours(inscolour)
        $frq.$panel tag config current -background $colours(textcur)\
                       -foreground $colours(textbg)
        # "blank" tag:
        $frq.$panel tag config curtitle    
        $frq.$panel tag config mancolour -background $colours(mancolour)
    }
    return
}

# KeyRight ... KeyDown --
#
#      One cell cursor motion routines.
#
proc KeyRight {} {
global sidecols cols c
    KeyForward [concat $sidecols $cols] $c "c"
    return
}

proc KeyLeft {} {
global sidecols cols c
    KeyBackward [concat $sidecols $cols] $c "c"
    return
}

proc KeyUp {} {
global headrows rows r
    KeyBackward [concat $headrows $rows] $r "r"
    return
}

proc KeyDown {} {
global headrows rows r
    KeyForward [concat $headrows $rows] $r "r"
    return
}

# UnPaintCurrentCell --
#
#      Remove "current cell" painting from WHOLE spreadsheet
#
proc UnPaintCurrentCell {} {
global headrows rows sidecols cols paintallowed
    if {$paintallowed == 1} {
        PaintCell "" "" "" nocurrent
        PaintCell "" "" "" nocurtitle
    }
    return
}

# PaintCurrentCell --
#
#      Paint the current cell with coloured tag "current", and tag
#      (for "see" operation, scrolling) associated row/col title cells
#
proc PaintCurrentCell {} {
global r c paintallowed
    if {$paintallowed == 1} {
        PaintCell [RowColPanel $r    $c] $r    $c current
        PaintCell [RowColPanel title $c] title $c curtitle
        PaintCell [RowColPanel $r title] $r title curtitle
    }
    return
}

# KeyForward, KeyBackward --
#
#      Service routines for cingle cell cursor motion keystrokes
#
# Arguments:
#      itemlist  List of rows or columns, depending rc
#      item      current row/col.
#      rc        Row (r) or column (c) motion.
#
proc KeyForward {itemlist item rc} {
global r c
    UnPaintCurrentCell
    set x [lindex $itemlist [Succ [lsearch $itemlist $item] [llength $itemlist] ] ]
    if {$rc == "r"} {set r $x} {set c $x}
    PaintCurrentCell
    return
}

proc KeyBackward {itemlist item rc} {
global r c
    UnPaintCurrentCell
    set x [lindex $itemlist [Pred [lsearch $itemlist $item] ] ]
    if {$rc == "r"} {set r $x} {set c $x}
    PaintCurrentCell
    return
}

# Put an info label stating default dir of motion, autocalc:
#
proc PutInfo {} {
global defmov tools frq
    switch -- $defmov {
        {-}     {} 
        {<} KeyLeft
        {>} KeyRight 
        {^} KeyUp
        {v} KeyDown
    }
    $frq.flag.mv configure -text $defmov -font fixed
    return
}

#
# Keep track of focus status:
#
proc NavFocus {} {
global frq colours
    focus $frq
    $frq.flag.flag configure -text "NAV" -foreground $colours(mancolour)
    .w config -cursor {}
    return
}

proc EditFocus {} {
global frq colours
    focus .w.fr.box1.e
    $frq.flag.flag configure -text "EDIT" -foreground $colours(mancolour)
    return
}

#
# Keep track of modification status and display logo:
#
proc Modified {} {
global modified frq reload tools
    MakeLogo red
    set modified 1
    set reload 1
    if [winfo exists .w] {
        $tools.save config -state normal
    }
    return
}

proc UnModified {} {
global modified frq tools
    MakeLogo #00b000
    set modified 0
    if [winfo exists .w] {
        $tools.save config -state disabled
    }
    return
}

proc ContentMatch {} {
global rows headrows cols sidecols e f v rowsel colsel thisformula find
global r c
    if {$find(rowsel) == 1} {
        set prows $rowsel
    } else {
        set prows [ConcatNoTitle $headrows $rows] 
    }
    if {$find(colsel) == 1} {
        set pcols $colsel
    } else {
        set pcols [ConcatNoTitle $sidecols $cols]
    }
    set rowsel {}
    set colsel {}
    if $find(case) {set swit "--"} else {set swit "-nocase"}
    foreach x $prows {
        foreach y $pcols {
            if {$find(fv) == "Formula"} {
                 set z $f($x,$y)
            } else {set z $v($x,$y)}
            if [catch {regexp $swit [$e get] $z} result] {
                RollFormulaStackUp
                set thisformula $result
            } else {
                if $result {
                    if {[lsearch $rowsel $x] == -1} {
                        lappend rowsel $x
                    }
                lappend colsel $y
                }
            }
        }
    }
    PaintSelect
    return
}

# MakeLogo --
#
#      Draw the tk# logo in green or red (indicating "modified").
#
# Argument:
#      colour   either green or red
#
proc MakeLogo {colour} {
global colours
    return [if [info exists logo] {image delete logo} 
    image create bitmap logo -foreground $colour -data {
#define tk_pic2_width 30
#define tk_pic2_height 27
static unsigned char tk_pic2_bits[] = {
     0x00, 0x00, 0x00, 0x00, 0x10, 0x02, 0x00, 0x00, 0x10, 0x02, 0x00, 0x00,
     0x18, 0x03, 0x00, 0x00, 0x08, 0x11, 0x00, 0x00, 0xbe, 0x1d, 0x00, 0x00,
     0x88, 0x05, 0x00, 0x00, 0x8c, 0xe3, 0x0c, 0x03, 0x84, 0xe2, 0x0c, 0x03,
     0xc6, 0x06, 0x0c, 0x03, 0x42, 0x0c, 0x8e, 0x03, 0x7e, 0x08, 0x86, 0x01,
     0x00, 0x00, 0x86, 0x01, 0x00, 0xe0, 0xff, 0x0f, 0x00, 0xe0, 0xff, 0x0f,
     0x00, 0x00, 0xc3, 0x00, 0x00, 0x00, 0xc3, 0x00, 0x00, 0x80, 0x61, 0x00,
     0x00, 0x80, 0x61, 0x00, 0x00, 0xf8, 0xff, 0x03, 0x00, 0xf8, 0xff, 0x03,
     0x00, 0xc0, 0x30, 0x00, 0x00, 0xc0, 0x30, 0x00, 0x00, 0xe0, 0x38, 0x00,
     0x00, 0x60, 0x18, 0x00, 0x00, 0x60, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00};

    }] 
    return
}

proc RR {r1 r2} {
global rows headrows   
    set sw 0
    set y ""
    foreach x [ConcatNoTitle $headrows $rows] {
        if {$x == $r1} {set sw 1}
        if {$sw == 1} {set y "$y$x"}
        if {$x == $r2} {set sw 0}
        if {$sw == 1} {set y "$y|"}
    }
    return $y
}

# InsertFunction --
#
#      Service routine for inserting math functions in Formula box:
#
# Arguments:
#      be, en     The two halves of a function call to be inserted
#                 around the selection.
#
proc InsertFunction {be {en ""}} {
global e
    if [$e selection present] {
        $e insert sel.first $be    
        $e insert sel.last  $en
        $e selection from [expr [$e index sel.first] - [string length $be]]
        $e selection to   [expr [$e index sel.last]  + [string length $en]]
        
    } else {
        $e insert insert $be
        $e insert insert $en
        $e icursor [expr [$e index insert] - [string length $en]]
    }
    return
}

# ConcatNoTitle --
#
#      Concatenate two lists of row/col names, but leave "title" out
#
proc ConcatNoTitle {rr1 rr2} {
   set s {}
   set rr [concat $rr1 $rr2]
   foreach x $rr {
       if {$x != "title"} {lappend s $x}
   }
   return $s
}

proc SelectAllRows {} {
global headrows rows rowsel
    set rowsel [ConcatNoTitle $headrows $rows]
    PaintSelect
}


proc SelectAllCols {} {
global sidecols cols colsel
    set colsel [ConcatNoTitle $sidecols $cols]
    PaintSelect
}

proc SelectAll {} {
    SelectAllRows
    SelectAllCols
}

# BuildMenu --
#
#      Construct the pulldown menu of tk#.
#
proc BuildMenu {} {
global fr b exp lfl lfs defmov e reps r c find printcommand
global thisformula
    set b [frame $fr.box0 -relief raised -borderwidth 2]

    menubutton $b.file -text File -menu $b.file.menu -underline 0
    set m [menu $b.file.menu -tearoff 0]
        $m add command -label Load           -command {Load 1} 
        $m add command -label Save -command {Save 0} -accel ^S
        $m add command -label "Save As..."   -command {Save 1}
        $m add command -label "Save Select"  -command {Save 2}
	$m add separator
        $m add command -label "Load Repository" -command {Load 2} 
        $m add command -label "Show Repositories" -command {
            wm title .w "Rep: $reps"
            }
        $m add separator
        $m add cascade -label "Export/Import..." -menu $m.sub0
            set m2 [menu $m.sub0 -tearoff 1]
            $m2 add radio -label "All"       -variable exp(howmuch)
            $m2 add radio -label "Selection" -variable exp(howmuch)
            $m2 add radio -label "Row Sel"   -variable exp(howmuch)
            $m2 add radio -label "Col Sel"   -variable exp(howmuch)
            $m2 add separator
            $m2 add check -label "Title Row" -variable exp(titlerow)
            $m2 add check -label "Title Col" -variable exp(titlecol)
            $m2 add separator
            $m2 add radio -label "Tab"     -variable exp(separ)
            $m2 add radio -label "Comma"   -variable exp(separ)
            $m2 add radio -label "Blank"   -variable exp(separ)
            $m2 add radio -label "Newline" -variable exp(separ)
            $m2 add radio -label "None"    -variable exp(separ)
            $m2 add radio -label "LaTeX"   -variable exp(separ)
            $m2 add radio -label "LyX10"   -variable exp(separ) 
            $m2 add radio -label "LyX12"   -variable exp(separ)

            $m2 add cascade -label "LyX..." -menu $m2.sub0
                set m3 [menu $m2.sub0 -tearoff 1]
                $m3 add radio -label "Align Left"   -var exp(align)
                $m3 add radio -label "Align Right"  -var exp(align)
                $m3 add radio -label "Align Centre" -var exp(align)
            $m2 add separator
            $m2 add command -label "Export!" -command Export
            $m2 add command -label "Import!" -command Import

        $m add separator
        $m add command -label "|Pipe Selection" \
	     -accelerator "^P" -command  { PrintSel $rowsel $colsel } 
        $m add command -label "|Pipe All" \
          -command  {
	     PrintSel [concat $headrows $rows] [concat $sidecols $cols] 
	} 
        $m add separator
        $m add command -label Quit -command QuitWarn -accelerator "^Q"
        $m add separator
        # quotes to force eval here:
        $m add command -label "$lfl" -command "set filename $lfl ; Load 0"
        $m add command -label "$lfs" -command "set filename $lfs ; Load 0"

    menubutton $b.editrc -text Edit -menu .w.fr.box0.editrc.menu -underline 0
    set m [menu $b.editrc.menu -tearoff 0]
        $m add command -label "Row Spawn" -command {
            RCspawnWindow "Row Spawn" RowSpawn ".rs"
        }
        $m add command -label "Row Rename" -command {
            RCrenameWindow "Row Rename" "New Name Template:" RowRename\
                    .rren r
        } 
        $m add command -label "Row Move" -command RowMove 
        $m add command -label "Row Move & Sort" -command {RowMove 1}
        $m add command -label "Row Kill" -command RowKill 
        $m add separator
        $m add command -label "Col Spawn" -command {
            RCspawnWindow "Col Spawn" ColSpawn ".cs"
        }
        $m add command -label "Col Rename" -command {
            RCrenameWindow "Column Rename" "New Name Template:" ColRename\
                    .cren c
        } 
        $m add command -label "Col Move" -command ColMove 
        $m add command -label "Col Move & Sort" -command {ColMove 1}
        $m add command -label "Col Kill" -command ColKill 
        $m add separator
        $m add command -label "Make Sheet" -command {GetGeometry .geom}
        $m add command -label "Kill Sheet" -command {
            set choice \
            [tk_messageBox -type yesno -icon warning \
                -message "Kill Sheet: Are you sure?" -default no]
            if {$choice == "yes"} {
                SelectAll
                RowKill
                ColKill
                ReSizeHeaders
            }
        }
        $m add command -label "Kill Select" -command {
            set choice \
            [tk_messageBox -type yesno -icon warning \
                -message "Kill Select: Are you sure?" -default no]
            if {$choice == "yes"} {
                RowKill 
                ColKill
                ReSizeHeaders
            }
        }
        $m add separator
        $m add command -label "Add Row" -command AddRow 

    menubutton $b.sel -text Select -menu $b.sel.menu -underline 0
    set m [menu $b.sel.menu -tearoff 1]
        $m add command -label "Row Select by Listbox" -command {
            SelectEdit [ConcatNoTitle $headrows $rows] r \
                "Row Select by Listbox" multiple rowsel
        }
        $m add command -label "Column Select by Listbox" -command {
            SelectEdit [ConcatNoTitle $sidecols $cols] c \
                "Col Select by Listbox" multiple colsel
        } 
        $m add command -label "Title Match" -command SelectMatch

        $m add cascade -label "Content Match..." -menu $m.sub0
            set m2 [menu $m.sub0 -tearoff 1]
            $m2 add  check -label "Case Sensitive"  -var find(case)
            $m2 add separator
            $m2 add  radio -label "Formula"         -var find(fv)
            $m2 add  radio -label "Data"            -var find(fv)
            $m2 add separator
            $m2 add  check -label "Within Row Sel"  -var find(rowsel)
            $m2 add  check -label "Within Col Sel"  -var find(colsel)
            $m2 add separator
            $m2 add command -label "Match!" -acc "^V"\
                    -command ContentMatch

        $m add separator
        $m add command -label "Select All" -command SelectAll
        $m add command -label "Select All Rows" -command SelectAllRows
        $m add command -label "Select All Columns" -command SelectAllCols
        $m add separator
        $m add command -label "Set Insert Mark" -command {
            SelectEdit [concat $headrows $rows $sidecols $cols] i \
                 "Set Insert Mark" single inspoint
        }

    menubutton $b.format -text Format -menu $b.format.menu -underline 2
    set m [menu $b.format.menu -tearoff 1]
        $m add command -label "Fixed Real"    -acc "^F" -command ComposeFixed
        $m add command -label "Floating Real" -acc "^G" -command ComposeFloat
        $m add command -label "Integer"       -acc "^N" -command ComposeInt
        $m add command -label "Currency"      -acc "^C" -command ComposeCurr
        $m add command -label "Left-Align Text" -acc "^L"  -comm LeftAlignText
        $m add command -label "Left-Align"              -command LeftAlign
        $m add separator
        $m add command -label {Add        [Expr...]} -acc "^E" -command AddExpr
        $m add command -label {Remove [Expr...]} -command RemoveExpr
        $m add command -label {Remove [Fmt...]} -command RemoveFormat
        $m add separator
        $m add cascade -label "Act on Selection..." -menu $m.sub0
            set m2 [menu $m.sub0 -tearoff 1]
            $m2 add command -label "Fixed Real" \
                    -command {FormatSelect ComposeFixed}
            $m2 add command -label "Floating Real" \
                    -command {FormatSelect ComposeFloat}
            $m2 add command -label "Integer" \
                    -command {FormatSelect ComposeInt}
            $m2 add command -label "Currency" \
                    -command {FormatSelect ComposeCurr}
            $m2 add command -label "Left-Align Text" \
                    -command {FormatSelect LeftAlignText}
            $m2 add command -label "Left-Align" \
                    -command {FormatSelect LeftAlign}
            $m2 add separator
            $m2 add command -label {Add        [Expr...]} \
                    -command {FormatSelect AddExpr}
            $m2 add command -label {Remove [Expr...]} \
                    -command {FormatSelect RemoveExpr}
            $m2 add command -label {Remove [Fmt...]} \
                    -command {FormatSelect RemoveFormat}

        $m add separator
        $m add command -label "Replace by Value" \
            -command {set thisformula [subst [$e get]]} 
        $m add separator
        $m add command -label "Clear" -command {set thisformula ""} 

    menubutton $b.calc -text Options -menu $b.calc.menu -underline 0
    set m [menu $b.calc.menu -tearoff 0]
        $m add check -label AutoCalc -variable autocalc -command PutInfo
        $m add command -label ReCalc -command ReCalc
        $m add command -label "ReCalc Sel" -command {ReCalc 1}
        $m add command -label "ReDisplay"    -command ReDisplay
        $m add separator
        foreach x {- {<} {>} {^} {v} } {
            $m add radio -label $x -var defmov -font fixed -command PutInfo
        }
        set defmov "-"

    menubutton $b.math -text Math -menu $b.math.menu -underline 0
    set m [menu $b.math.menu -tearoff 0]
        $m add cascade -label "Functions 1..." -menu $m.sub1
            set m2 [menu $m.sub1 -tearoff 1]
        foreach x {acos cos sinh asin cosh log sqrt atan exp log10 tan tanh sin} {
            $m2 add command -label $x -command [list InsertFunction "$x\(" ")"]
        }
        $m add cascade -label "Functions 2..." -menu $m.sub2
            set m2 [menu $m.sub2 -tearoff 1]
        foreach x {floor ceil abs double int round hypot atan2 pow fmod} {
            $m2 add command -label $x -command [list InsertFunction "$x\(" ")"]
        }
        $m add cascade -label "Row Aggregate..." -menu $m.sub3
            set m2 [menu $m.sub3 -tearoff 1]
            foreach x {Sum Av Cnt Max Min wSum wAv Ssq sSsq RMS wRMS Sd wSd} {
                $m2 add command -label "$x" -acc "\[$x r regex]"\
                         -command [list InsertFunction "\[$x r " "\]"]
            }
            $m2 add separator
            # Here a match string for selected set of rows/cols is inserted
            # (Thanks Bruce Gingery for the idea):
            $m2 add command -label "Matchmaker" -acc " ^(c1|c2|c3)$" -command {
                $e insert insert "^("
                $e insert insert [join $colsel |]
                $e insert insert ")$"
                # necessary :-(
                set colsel $c
            }
            #Bruce Gingery and Juergen Vigna cannot both be wrong...
            $m2 add command -label "Column Range" -acc "   \[CC c1 c2]"\
                     -command [list InsertFunction "\[CC " "\]"]
            $m2 add command -label "This Column"  -acc "   \$c"\
                     -command {InsertFunction "$c "}

        $m add cascade -label "Col Aggregate..." -menu $m.sub4
            set m2 [menu $m.sub4 -tearoff 1]
            foreach x {Sum Av Cnt Max Min wSum wAv Ssq sSsq RMS wRMS Sd wSd} {
                $m2 add command -label "$x" -acc "\[$x c regex]"\
                        -command [list InsertFunction "\[$x c " "\]"]
            }
            $m2 add command -label "Matchmaker" -acc " ^(r1|r2|r3)$" -command {
                $e insert insert "^("
                $e insert insert [join $rowsel |]
                $e insert insert ")$"
                set rowsel $r
            }
            $m2 add command -label "Row Range" -acc "   \[RR r1 r2]"\
                       -command [list InsertFunction "\[RR " "\]"]
            $m2 add command -label "This Row"  -acc "   \$r"\
                       -command {InsertFunction "$r "}
        
        $m add separator

        $m add command -label "This Row" -acc "\$r"\
                 -command {InsertFunction "$r "}
        $m add command -label "This Column" -acc "\$c"\
                 -command {InsertFunction "$c "}


    menubutton $b.clock -text Clock -menu $b.clock.menu -underline 0
    set m [menu $b.clock.menu -tearoff 0]
        $m add command -label "(Re-)Start" -command StartClock
        $m add command -label "Stop"       -command StopClock

    menubutton $b.colour -text Colour -menu $b.colour.menu -underline 2
    set m [menu $b.colour.menu -tearoff 1]
        $m add command -lab "Foreground"   -command {ColourReconfig textfg}
        $m add command -lab "Background"   -command {ColourReconfig textbg} 
        $m add command -lab "Current Cell" -command {ColourReconfig textcur} 
        $m add command -lab "Insert Point" -command {ColourReconfig inscolour} 
        $m add command -lab "Selection"    -command {ColourReconfig selcolour} 
        $m add command -lab "Clock Cell"   -command {ColourReconfig timecolour} 
        $m add separator
        $m add command -lab "tk# icon"     -command {ColourReconfig iconcolour} 
        $m add command -lab "man titles"   -command {ColourReconfig mancolour} 
        $m add separator
        $m add command -lab "Default"      -command {
            ColourDefault
            ColoursSet
        } 

    menubutton $b.help -text Help -menu $fr.box0.help.menu -underline 0 
    set m [menu $b.help.menu -tearoff 0]
        $m add command -label "Manual"   -command Manual 
        $m add command -label "About..." -command About
}

#
# ShowWindow --
# 
#      Construct and display rest of the GUI of tk#.
#
proc ShowWindow {} {
global thisformula cwidth titlerow titlecol
global colours winsize find defmov fonts
global filename autocalc inspoint headrows rows sidecols cols
#
# Symbolic widget names:
#      b     menubutton row
#      e     Formula:  entrybox
#      ef    Format:   entrybox
#      ew    Colwidth: entrybox
#      fr    frame around everything
#      frq   frame around spreadsheet panels
#      tools tool button bar.
#
global b e ef ew fr frq tools
#
    if {[winfo exists .w]} { wm deiconify .w; return }
    #
    # Widgets definition: 
    #
    wm withdraw .
    toplevel .w
    wm geometry  .w $winsize
    wm minsize   .w 1 1
    wm resizable .w 1 1
    wm title     .w "tk# -- a spreadsheet with a difference"

    bind .w <Destroy> {if {"%W" == ".w"} {QuitWarn 1}}
    bind .w <Control-s> {Save 0}
    bind .w <Control-q> QuitWarn
    bind .w <Control-p> { PrintSel $rowsel $colsel }

    set fr  [frame .w.fr -borderwidth 2]
    set frq [frame $fr.frq -highlightcolor black -highlightthickness 2 \
                     -borderwidth 0]

    set logo [MakeLogo #00b000]

    button $frq.logo -padx 0 -pady 0 -image logo \
         -relief raised -borderwidth 2 -command {Save 0}
    frame $frq.flag
    label $frq.flag.flag -anchor w -pady 0
    label $frq.flag.mv -text "-" -foreground $colours(mancolour)
    pack  $frq.flag.flag -side left
    pack  $frq.flag.mv   -side right
    bind  $frq.flag <Button> NavFocus
    NavFocus

    set tools [frame $fr.tools -borderwidth 0]

    # toolbar:
    button $tools.save  -text Save  -command {Save 0} -state disabled
    button $tools.calc  -text Calc  -command ReCalc
    button $tools.view  -text View  -command {
        set paintallowed 1
        MenusState normal
        ReDisplay
        PutInfo
    }
    button $tools.start -text Start -command StartClock
    button $tools.stop  -text Stop  -command StopClock -state disabled
    button $tools.add   -text Add   -command AddRow
    button $tools.find  -text Find  -command {
        set find(case) 0
        set find(fv) "Data"
        ContentMatch
    } 

    button $tools.export  -text Export   -command Export

   checkbutton  $tools.auto  -text "Auto" -width 5 -variable autocalc\
               -anchor w -foreground $colours(mancolour)
    label  $tools.pad1 -width 1 
    label  $tools.pad2 -width 1
    label  $tools.pad3 -width 1
    PutInfo

    foreach x {save calc view start stop add find export auto pad1 pad2 pad3} {
        $tools.$x config -pady 0 
    }

    # The two scrollbars:
    scrollbar $frq.hscroll -borderwidth 1 -comm {HorizScroll} -orient horiz 
    scrollbar $frq.vscroll -borderwidth 1 -comm {VertScroll}  -orient vert 

    # The four main spreadsheet panels:
    text $frq.main -wrap none -font $fonts(mono) -background $colours(textbg) \
                     -foreground $colours(textfg)\
                -xscrollcommand {.w.fr.frq.hscroll set} \
                -yscrollcommand {.w.fr.frq.vscroll set}
    text $frq.head -wrap none -background $colours(textbg) \
                 -foreground $colours(textfg) -font $fonts(mono)\
                 -height 3 -xscrollcommand {.w.fr.frq.hscroll set} 
    text $frq.side -wrap none -background $colours(textbg) \
                 -foreground $colours(textfg)\
                 -font $fonts(mono) -width 20 -yscrollcommand {.w.fr.frq.vscroll set} 
    text $frq.corner -width 20 -height 3 -font $fonts(mono) -wrap none\
                -background $colours(textbg) -foreground $colours(textfg) 

    BuildMenu

    frame $fr.box1
    menubutton $fr.box1.l -text "Formula:" -menu $fr.box1.l.menu 
    set m [menu $fr.box1.l.menu -tearoff 1]
         $m add comm -label "Scroll Up"     -acc "^<Up>"   \
                 -comm RollFormulaStackUp
         $m add comm -label "Scroll Down  " -acc "^<Down>" \
                 -comm RollFormulaStackDown
         $m add comm -label "Writeback"  -acc "<Enter>"  -comm WriteFormula
         $m add comm -label "Clear"         -acc "^<Del>"  -comm {
             set thisformula ""
         }
         $m add comm -label "Load Cell"  -acc "^<Enter>" -comm ReadFormula
         $m add comm -label "Toggle NAV/EDIT" -acc "<Esc>"  -comm NavFocus

    set e [entry $fr.box1.e -textvariable thisformula \
                 -background $colours(textbg)]
    bind $e <Key-Return> WriteFormula
    bind $e <Control-Return> {
        NavFocus
        ReadFormula
    }

    bind $e <Key-Escape> NavFocus
    bind $e <Key-Down> {NavFocus; KeyDown}
    bind $e <Key-Up>   {NavFocus; KeyUp  }

    bind $e <Button> EditFocus

    # Keybindings for entry formatting:
    bind $e <Control-f> ComposeFixed
    bind $e <Control-g> ComposeFloat
    bind $e <Control-n> ComposeInt
    bind $e <Control-c> ComposeCurr
    bind $e <Control-e> AddExpr
    bind $e <Control-l> LeftAlign

    bind $e <Control-Key-Up>   RollFormulaStackUp
    bind $e <Control-Key-Down> RollFormulaStackDown
    bind $e <Control-Key-Delete> {set thisformula ""} 

    bind $e <Control-v> ContentMatch

    label $fr.box1.lw -text "Colwidth:" -width 8 -anchor e
    set ew [entry $fr.box1.ew -textvariable cwidth \
                -background $colours(textbg) -width 4]
    bind $ew <Key-Return> ColWidth
    bind $ew <Key-Escape> NavFocus
    bind $ew <Button>     EditFocus

    label $fr.box1.lf -text "Format:" -width 7 -anchor e
    set ef [entry $fr.box1.ef -textvariable fmtitem \
                -background $colours(textbg) -width 4]
    bind $ef <Button>     EditFocus
    bind $ef <Key-Escape> NavFocus

    foreach panel {main head side corner} {
        $frq.$panel config -cursor arrow
        bind $frq.$panel <Button-1>         [list GetMouse1 $panel 1]
        bind $frq.$panel <Control-Button-1> [list GetMouse1 $panel 0] 
        bind $frq.$panel <Button-3>         [list GetMouse3 $panel] 
    }
    
    # bind the middle mouse button dragging function:
    foreach panel {main head side corner} {
        bind $frq.$panel <B2-Motion>\
              [list BindDragto %x %y $frq.main $frq.head $frq.side $frq.corner]
        bind $frq.$panel <Button-2> {
              [list BindMark   %x %y $frq.main $frq.head $frq.side $frq.corner]
        }
    } 

    foreach panel {main head side corner} {
        bind $frq.$panel <Button-2> [list GetMouse2 $panel]
    }

    # The arrow keys:
    bind $frq <Key-Right> KeyRight
    bind $frq <Key-Left>  KeyLeft
    bind $frq <Key-Down>  KeyDown
    bind $frq <Key-Up>    KeyUp

    # Ctrl-arrow keys scroll one pageful at a time (co-ordinatedly!):
    bind $frq <Control-Key-Right> {
        $frq.main xview scroll  1 pages
        $frq.head xview scroll  1 pages
    }
    bind $frq <Control-Key-Left>  {
        $frq.main xview scroll -1 pages
        $frq.head xview scroll -1 pages
    }
    bind $frq <Control-Key-Down>  {
        $frq.main yview scroll  1 pages
        $frq.side yview scroll  1 pages
    }
    bind $frq <Control-Key-Up>    {
        $frq.main yview scroll -1 pages
        $frq.side yview scroll -1 pages
    }
 
    bind $frq <Control-Return> ReadFormula
    # Make selection = current cell:
    bind $frq <Key-Return> {
        SetUpForEdit
        NavFocus
        }
    bind $frq <Key-Escape> EditFocus
    # Empty edit box:
    bind $frq <Key-Delete> {
        RollFormulaStackUp
        set thisformula "" 
        EditFocus
    }
    # Combine Return and Delete effects:
    bind $frq <Control-Delete> {
        SetUpForEdit
        RollFormulaStackUp
        set thisformula "" 
        EditFocus
    }

    foreach panel {side head corner} {
        $frq.$panel tag config underline -background #d9d9d9 \
                -relief raised -borderwidth 2
    }

    TagsConfig
 
    #
    # Geometry definition: 
    #

    grid columnconf .w 0 -weight 1
    grid rowconf    .w 0 -weight 1

    grid $fr -in .w -column 0 -row 0 -sticky news 

    grid columnconf $fr 0 -weight 1
    grid rowconf    $fr 3 -weight 1

    pack $b.file $b.editrc $b.format $b.sel \
             $b.calc $b.math $b.colour $b.clock -side left
    pack $b.help -side right

    pack $fr.box1.l -side left
    pack $fr.box1.e -side left -fill x -expand true
    pack $fr.box1.ew $fr.box1.lw $fr.box1.ef $fr.box1.lf -side right

    grid $b       -sticky ew
    grid $fr.box1 -sticky ew

    grid $tools   -sticky ew

    pack $tools.auto $tools.save\
            $tools.pad1 $tools.calc $tools.view $tools.find \
            $tools.pad2 $tools.start $tools.stop $tools.pad3 $tools.add\
            $tools.export -side left

    grid $frq -in $fr -sticky news

    grid columnconf $frq 2 -weight 1
    grid rowconf    $frq 2 -weight 1

    grid $frq.logo    $frq.flag   $frq.hscroll -sticky new
    grid ^            $frq.corner $frq.head    -sticky news
    grid $frq.vscroll $frq.side   $frq.main    -sticky news

    focus $frq
    UnModified
    return
}

init $argc $argv

ShowWindow 

# command line filename(s):
foreach filename $argv {
    Load 0
}
 
if {$argc == 0} {
    # No command line file was loaded:
    ReDisplay
    GetGeometry .geom
}
